"""Unit-test for get/set/injection in Evaluator

Unit-test for assignment, modification, getting, and variable

Run `python3 -m pytest tests/test_evaluator_get_set.py` in
parent directory.

Environment:
- Python 3.9 or later
- pytest 8.3 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-23'

# pylint: disable=C0301, E0401, R0801

import pytest
from vivjson.viv import Viv
from vivjson.config import Config

@pytest.mark.parametrize(
    'prepared_expression, target_expression, expected_value, is_error', (
        ({"a": 3}, 'a', 3, False),
        ({"a": NotImplemented}, 'a', None, True),  # Python's NotImplemented can't treated. --> ERROR
        ({"a": Config()}, 'a', None, True),  # Python's class instance can't treated. --> ERROR
        ({"a": "x", "b": "y"}, 'a', "x", False),
        ('a = 1.5', 'a', 1.5, False),
        ('a = 1.5, a+=0.5', 'a', 2.0, False),
        ('a = 3 > 2', 'a', True, False),
        ('a = 3 <= 2', 'a', False, False),
        ('a = "xyz", b = a == "xyz"', 'b', True, False),
        ('a = "xyz", b = a != "xyz"', 'b', False, False),
        ('a = {"x": 100}', 'a.x', 100, False),
        ('a = {"x": 100}', 'a.y', None, False),
        ('a = {"x": 100, "_y": 20}', 'a.x', 100, False),
        ('a = {x: 10}', 'a.x', 10, False),
        ('a: {x: -1e4}', 'a["x"]', -10000.0, False),
        ('"a": {"x": "true", y: "false"}', "a['x'] + '&' + a['y']", "true&false", False),
        ('a = [100 10 20]', 'a[0] + a[1] + a[2]', 130, False),
        ('a = [100 10 20] a[1] += 1000', 'a[0] + a[1] + a[2]', 1130, False),
        ('a = [1 5 {b: {c: 10} d: -20}]', 'a.2.b.c', 10, False),
        ('a = [1 5 {b: {c: 10} d: -20}]', 'a[2]["b"]["c"]', 10, False),
        ('a = [1 5 {b: {c: 10} d: -20}]', 'a[2].b["c"]', 10, False),
        ('a = [1 5 {b: {c: 10} d: -20}]', 'a[-1].b.c', 10, False),
        ('', 'a', None, False),  # The undefined variable's value is null.
        ('', 'a.b', None, False),  # The member's value of the undefined variable is null.
        ('', 'a["b"]', None, False),  # The member's value of the undefined variable is null.
        ('', 'a[1]', None, False),  # The member's value of the undefined variable is null.
        ('a = {}', 'a', {}, False),
        ('a = []', 'a', [], False),
        ('a = [true, 100]', 'a[0]', True, False),
        ('a = [true, 100]', 'a[1]', 100, False),
        ('a = [true, 100]', 'a[2]', None, False),  # null is gotten even if out of range
        ('a = [true, 100]', 'a[-1]', 100, False),
        ('a = [true, 100]', 'a[-2]', True, False),
        ('a = [true, 100]', 'a[-3]', None, False),  # null is gotten even if out of range
        ('a = [true, 100]', 'a[false]', None, True),  # boolean can not be used as index.
        ('a = [true, 100], a[0] = 20', 'a[0]', 20, False),
        ('a = [true, 100], x = 0, a[x] = 10', 'a[0]', 10, False),
        ('a = [true, 100], x = 0, a[x] = 10', 'a[x]', 10, False),
        ('a = [true, 100], x = 0', 'a.x', None, True),  # a.x is invalid.
        ('a = [true, 100], b = 2.5 * 0.4', 'a[b]', 100, False),
        ('a = [true, 100]', 'a[2.5 * 0.4]', 100, False),
        ('a = [true, 100]', 'a[2.5 * 0.3]', None, True),  # 2.5 * 0.3 = 0.75 is invalid as index.
        ('a = [true, 100], a[2] = 20', 'a[0]', None, True),  # a[2] makes error.
        ('a = [true, 100], a += [false, 20]', 'a', [True, 100, [False, 20]], False),
        ('a = {x: true y: 100} a += {x: false, y: 20}', 'a', {"x": True, "y": 120}, False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x["b"][0]', True, False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x["b"].0', True, False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b[0]', True, False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b.0', True, False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b.-2', True, False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b[-2]', True, False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b["0"]', None, True),  # The index of array must be number.
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b.1.0', "foo", False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b.1["0"]', "foo", False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b.1[0]', "foo", False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b.-1[0]', "foo", False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b[-1].0', "foo", False),
        ('x = {"a": 100, "b": [true, {"c": 3, "0": "foo"}]}', 'x.b.-1.0', "foo", False),
        ('a = {b: 3, c: {d: true, e: [null, "x", false]}}', 'a.c.e.1', "x", False),
        ('a = {b: 3, c: {d: true, e: [null, "x", false]}}', 'a["c"].e[-2]', "x", False),
        ('b = 1, a = {b: 3, c: {d: true, e: [null, "x", false]}}', 'a["c"].e[b]', "x", False),
        ('a = {b: 1, "c": 2}; a.d = 3', 'a["d"]', 3, False),
        ('a = [{"b": 1, c: 2}, 3], a[0].b = 5', 'a.0["b"]', 5, False),
        ('a = [{"b": 1, c: 2}, 3], a[0].b = 5', 'a.0. b', 5, False),
        ('a = [{"b": 1, c: 2}, 3], a[0].b = 5', 'a.0.*', None, True),  # .* is invalid.
        ('a = [{"b": 1, c: 2}, 3], a[0].b = 5', 'a.0. *', None, True),  # .* is invalid.
        ('a = {"0": 100, "1": 50}', 'a.0', 100, False),
        ('a = {"0": 100, "1": 50}, a.0 += 80', 'a.0', 180, False),
        ('a = {"0": [10, 30], "1": 50}, a.0 = 80', 'a.0', 80, False),
        ('a = {"0": [10, 30], "1": 50}, a.0.1 += 80', 'a.0[1]', 110, False),
        ('a = {"0": [10, 30], "1": 50}, a.0.1 += 80', 'a.0.1', 110, False),
        ('a = {"0": [10, [30, 70]], "1": 50}, a.0.1.1 += 80', 'a.0.1', [30, 150], False),
        ('a = [1, 2]', 'a[0.5]', None, True),  # Index must be integer.
        ('a = [1, 2]', 'a[1.0]', 2, False),
        ('a = [1, 2], a[0.5] = 1.5', 'a', None, True),  # Index must be integer.
        ('a = [1, 2], a[1.0] = 1.5', 'a', [1, 1.5], False),
        ('x = 10, y = if (x >= 0) { s = "+" } else { s = "-" }', 'y', {"s": "+"}, False),
        ('function x2(k) {return(_[0] * 2)}', 'x2(3)', 6, False),
        ('function x2(k) {_[0] *= 2 return(k)}', 'x2(3)', 6, False),
        ('function x2(k) {return(k * 2)}', 'x2(3)', 6, False),
        ('function x2(k) {k *= 2 return(_[0])}', 'x2(3)', 6, False),
        ('a = {b: 3}, function x2(k) {k.b *= 2 return(k.b)}', 'x2(a)', 6, False),
        ('a = {b: 3}, function x2(k) {k.b *= 2 return(_[0]["b"])}', 'x2(a)', 6, False),
        ('a = {b: 3}, function x2(k) {_[0]["b"] *= 2 return(k.b)}', 'x2(a)', 6, False),
        ('a = {b: 3}, function x2(k) {k.b *= 2 return(k.b)}, x2(a)', 'a.b', 3, False),
        ('a = {b: 3}, function x2(k) {return(k.b * 2)}', 'x2(a)', 6, False),
        ('while = 3, x = 2, while(x < 10) {x+=10}', '[while, x]', [3, 12], False),  # "while" is not reserved word.
        ('function while(a) { return(true) }', 'while(10)', True, False),  # "while" is not reserved word.
        ('for = 3', 'for', None, True),  # "for" is reserved word.
        ('import = 3', 'import', None, True),  # "import" is reserved word.
        ('super = 3', 'super', None, True),  # "super" is reserved word.
        ('___#RESULT#___ = 1', '___#RESULT#___', None, True),  # "RESULT#___ = 1" is ignored as comment.
        ('___i___ = 1', '___i___', None, True),  # It is invalid that both of prefix and suffix are "___".
        ('______ = 1', '______', None, True),  # It is invalid that both of prefix and suffix are "___".
        ('_____ = 1', '_____', 1, False),
        ('a = {b: 3, c:[2, 1]}, remove(a)', 'a', None, False),
        ('a = {b: 3, c:[2, 1]}, remove(a.b)', 'a', {"c": [2, 1]}, False),
        ('a = {b: 3, c:[2, 1]}, remove(a.c)', 'a', {"b": 3}, False),
        ('a = {b: 3, c:[2, 1]}, remove(a["b"])', 'a', {"c": [2, 1]}, False),
        ('a = {b: 3, c:[2, 1]}, remove(a.c.0)', 'a', {"b": 3, "c": [1]}, False),
        ('a = {b: 3, c:[2, 1]}, remove(a.c.1)', 'a', {"b": 3, "c": [2]}, False),
        ('a = {b: 3, c:[2, 1]}, remove(a.c.-2)', 'a', {"b": 3, "c": [1]}, False),
        ('a = {b: 3, c:[2, 1]}, remove(a.c.-1)', 'a', {"b": 3, "c": [2]}, False),
        ('a = {b: 3, c:[2, 1]}, remove(a.c[-1])', 'a', {"b": 3, "c": [2]}, False),
        ('a = {b: 3, c:[2, 1]}, remove(a["c"][-1])', 'a', {"b": 3, "c": [2]}, False),
        ('a = {b: 3, c:[2, 1]}, remove(a["c"])', 'a', {"b": 3}, False),
        ('a = {b: 3, c:[2, 1]}, remove(a.c.2)', 'a', None, True),  # Out of range
        ('a = {b: 3, c:[2, 1]}, remove(a["c"][2])', 'a', None, True),  # Out of range
        ('a = {b: 3, c:[2, 1]}, remove(a["c"][-3])', 'a', None, True),  # Out of range
        ('a = {b: 3, c:[2, 1]}, remove(a.d)', 'a', {"b": 3, "c":[2, 1]}, False),  # Nothing
        ('a = {b: 3, c:[2, 1], d: {e: false, remove(c[1])}}', 'a', {"b": 3, "c":[2], "d": {"e": False}}, False),
        ('a = 100, do {a: 100, remove(a)}', 'a', 100, False),
        ('a = 100, do {remove(a)}', 'a', None, False),
        ('do {remove(a)}', 'a', None, False),
        ('a={x:[30, 2]}; b=a; b.x.1=5; remove(b.x.0)', '[a, b]', [{"x": [30, 2]}, {"x": [5]}], False),
    )
)

def test_get_set(prepared_expression, target_expression, expected_value,
                 is_error):
    """Unit-test for get/set expression in Evaluator

    Unit-test for assignment, modification, getting, and variable

    Args:
        prepared_expression (str or dict): When the another expression,
                                 such as assignment, is needed before
                                 test, it is given here.
                                 Maybe this is '' mostly.
                                 str: Statements as VivJson
                                 dict: Value as Python's value
        target_expression (str): "return" method's argument as targeted
                                 expression.
        expected_value (Any): The expected value as Python's value
        is_error (bool): True if occurring error is expected,
                         False otherwise.
    """
    config = Config(max_depth=10)

    text = f'return({target_expression})'
    value, error_message = Viv.run(prepared_expression, text, config)
    assert (error_message != '') == is_error

    if isinstance(value, float):
        value = round(value, 10)
    assert value == expected_value

    if error_message == '' \
            and not isinstance(expected_value, bool) \
            and isinstance(expected_value, (int, float)):
        assert type(value) == type(expected_value)  # pylint: disable=C0123



"""Unit-test for order of operations in Evaluator

Unit-test for +, -, *, /, and, or, ...

Run `python3 -m pytest tests/test_evaluator_order.py` in
parent directory.

Environment:
- Python 3.9 or later
- pytest 8.3 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-20'

# pylint: disable=C0301, E0401, R0801

import pytest
from vivjson.viv import Viv
from vivjson.config import Config

@pytest.mark.parametrize(
    'prepared_expression, target_expression, expected_value, is_error', (
        ('', '1 + 2 * 5', 11, False),
        ('', '(1 + 2) * 5', 15, False),
        ('', '2 * 3 + 8', 14, False),
        ('', '3 / 5 + 10', 10.6, False),
        ('', '3 / (5 + 3 + 3 + 3 + 1)', 0.2, False),
        ('', '12 % 5 + 1', 3, False),
        ('', '12 % (5 + 1)', 0, False),
        ('', '3 / 5 * 3', 1.8, False),
        ('', '3 / (5 * 3)', 0.2, False),
        ('', '3 / (4 + 5 + 6)', 0.2, False),
        ('', '6 / ((4 + 5 + 6))', 0.4, False),
        ('', '(6 / ((4 + 5 + 6)))', 0.4, False),
        ('', '1 + (-6 / (4 + 5 + 6))', 0.6, False),
        ('', '-1 / 10', -0.1, False),
        ('', '3 +-1 / 10', 2.9, False),
        ('', '3 /-1', -3, False),
        ('', '("a,b" + "c") / ","', ["a", "bc"], False),
        ('', '"a,b" + "c" / ","', ["a,b", "c"], False),
        ('', '3 + [] + true', [3, True], False),
        ('', '3 + ["x"] + true', [3, "x", True], False),
        ('', 'false or true and true or false', True, False),
        ('', 'false or not false', True, False),
        ('', 'not false or false', True, False),
        ('', 'not true and true', False, False),
        ('', 'not(true and true)', False, False),
        ('', 'not (3 != 3)', True, False),
        ('', 'not (true)', False, False),
        ('', '(true)', True, False),
        ('', '3 > 2 and 0 > -1', True, False),
        ('', '3 > (2 and 0) > -1', None, True),  # 3 > false > -1 is invalid.
        ('', '3 > 2 and 2 > -1', True, False),
        ('a = 5; a += 3 * 2', 'a', 11, False),
        ('', '"a" and "b" == "c"', False, False),  # --> true and false --> false
        ('', '("a" and "b") == "c"', True, False),  # --> true == true --> true
    )
)

def test_order(prepared_expression, target_expression, expected_value,
               is_error):
    """Unit-test for order of operations in Evaluator

    Unit-test for +, -, *, /, and, or, ...

    Args:
        prepared_expression (str): When the another expression,
                                 such as assignment, is needed before
                                 test, it is given here.
                                 Maybe this is '' mostly.
        target_expression (str): "return" method's argument as targeted
                                 expression.
        expected_value (Any): The expected value as Python's value
        is_error (bool): True if occurring error is expected,
                         False otherwise.
    """
    config = Config(max_depth=10)

    text = f'{prepared_expression}; return({target_expression})'
    value, error_message = Viv.run(text, config)

    assert (error_message != '') == is_error

    if isinstance(value, float):
        value = round(value, 10)
    assert value == expected_value

    if error_message == '' \
            and not isinstance(expected_value, bool) \
            and isinstance(expected_value, (int, float)):
        assert type(value) == type(expected_value)  # pylint: disable=C0123



"""Unit-test for Config

Unit-test for getter/setter in Config

Run `python3 -m pytest tests/test_config.py` in parent directory.

Environment:
- Python 3.9 or later
- pytest 8.3 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-25'

# pylint: disable=C0121, E0401

from vivjson.config import Config

def test_config():  # pylint: disable=R0915
    """Unit-test for Config

    Unit-test for getter/setter in Config
    """
    assert Config.get_enable_stderr_alt(None) == Config.ENABLE_STDERR_DEFAULT
    assert Config.get_enable_tag_detail_alt(None) == Config.ENABLE_TAG_DETAIL_DEFAULT
    assert Config.get_enable_only_json_alt(None) == Config.ENABLE_ONLY_JSON_DEFAULT
    assert Config.get_infinity_alt(None) == Config.INFINITY_DEFAULT
    assert Config.get_nan_alt(None) == Config.NAN_DEFAULT
    assert Config.get_max_array_size_alt(None) == Config.MAX_ARRAY_SIZE_DEFAULT
    assert Config.get_max_depth_alt(None) == Config.MAX_DEPTH_DEFAULT
    assert Config.get_max_loop_times_alt(None) == Config.MAX_LOOP_TIMES_DEFAULT

    config = Config()

    assert config.get_enable_stderr() == Config.ENABLE_STDERR_DEFAULT
    assert Config.get_enable_stderr_alt(config) == Config.ENABLE_STDERR_DEFAULT
    assert config.get_enable_tag_detail() == Config.ENABLE_TAG_DETAIL_DEFAULT
    assert Config.get_enable_tag_detail_alt(config) == Config.ENABLE_TAG_DETAIL_DEFAULT
    assert config.get_enable_only_json() == Config.ENABLE_ONLY_JSON_DEFAULT
    assert Config.get_enable_only_json_alt(config) == Config.ENABLE_ONLY_JSON_DEFAULT
    assert config.get_infinity() == Config.INFINITY_DEFAULT
    assert Config.get_infinity_alt(config) == Config.INFINITY_DEFAULT
    assert config.get_nan() == Config.NAN_DEFAULT
    assert Config.get_nan_alt(config) == Config.NAN_DEFAULT
    assert config.get_max_array_size() == Config.MAX_ARRAY_SIZE_DEFAULT
    assert Config.get_max_array_size_alt(config) == Config.MAX_ARRAY_SIZE_DEFAULT
    assert config.get_max_depth() == Config.MAX_DEPTH_DEFAULT
    assert Config.get_max_depth_alt(config) == Config.MAX_DEPTH_DEFAULT
    assert config.get_max_loop_times() == Config.MAX_LOOP_TIMES_DEFAULT
    assert Config.get_max_loop_times_alt(config) == Config.MAX_LOOP_TIMES_DEFAULT

    config = Config(
        enable_stderr=True,
        enable_tag_detail=True,
        enable_only_json=True,
        infinity='Infinity',
        nan='NaN',
        max_array_size=3,
        max_depth=2,
        max_loop_times=1
    )

    assert config.get_enable_stderr() == True
    assert Config.get_enable_stderr_alt(config) == True
    assert Config.get_enable_stderr_alt(None) == Config.ENABLE_STDERR_DEFAULT
    assert config.get_enable_tag_detail() == True
    assert Config.get_enable_tag_detail_alt(config) == True
    assert Config.get_enable_tag_detail_alt(None) == Config.ENABLE_TAG_DETAIL_DEFAULT
    assert config.get_enable_only_json() == True
    assert Config.get_enable_only_json_alt(config) == True
    assert Config.get_enable_only_json_alt(None) == Config.ENABLE_ONLY_JSON_DEFAULT
    assert config.get_infinity() == 'Infinity'
    assert Config.get_infinity_alt(config) == 'Infinity'
    assert Config.get_infinity_alt(None) == Config.INFINITY_DEFAULT
    assert config.get_nan() == 'NaN'
    assert Config.get_nan_alt(config) == 'NaN'
    assert Config.get_nan_alt(None) == Config.NAN_DEFAULT
    assert config.get_max_array_size() == 3
    assert Config.get_max_array_size_alt(config) == 3
    assert Config.get_max_array_size_alt(None) == Config.MAX_ARRAY_SIZE_DEFAULT
    assert config.get_max_depth() == 2
    assert Config.get_max_depth_alt(config) == 2
    assert Config.get_max_depth_alt(None) == Config.MAX_DEPTH_DEFAULT
    assert config.get_max_loop_times() == 1
    assert Config.get_max_loop_times_alt(config) == 1
    assert Config.get_max_loop_times_alt(None) == Config.MAX_LOOP_TIMES_DEFAULT

    config = Config()

    assert config.get_enable_stderr() == Config.ENABLE_STDERR_DEFAULT
    assert Config.get_enable_stderr_alt(config) == Config.ENABLE_STDERR_DEFAULT
    assert Config.get_enable_stderr_alt(None) == Config.ENABLE_STDERR_DEFAULT
    config.enable_stderr()
    assert config.get_enable_stderr() == True
    assert Config.get_enable_stderr_alt(config) == True
    assert Config.get_enable_stderr_alt(None) == Config.ENABLE_STDERR_DEFAULT
    config.enable_stderr(False)
    assert config.get_enable_stderr() == False
    assert Config.get_enable_stderr_alt(config) == False
    assert Config.get_enable_stderr_alt(None) == Config.ENABLE_STDERR_DEFAULT
    config.enable_stderr(True)
    assert config.get_enable_stderr() == True
    assert Config.get_enable_stderr_alt(config) == True
    assert Config.get_enable_stderr_alt(None) == Config.ENABLE_STDERR_DEFAULT

    assert config.get_enable_tag_detail() == Config.ENABLE_TAG_DETAIL_DEFAULT
    assert Config.get_enable_tag_detail_alt(config) == Config.ENABLE_TAG_DETAIL_DEFAULT
    assert Config.get_enable_tag_detail_alt(None) == Config.ENABLE_TAG_DETAIL_DEFAULT
    config.enable_tag_detail()
    assert config.get_enable_tag_detail() == True
    assert Config.get_enable_tag_detail_alt(config) == True
    assert Config.get_enable_tag_detail_alt(None) == Config.ENABLE_TAG_DETAIL_DEFAULT
    config.enable_tag_detail(False)
    assert config.get_enable_tag_detail() == False
    assert Config.get_enable_tag_detail_alt(config) == False
    assert Config.get_enable_tag_detail_alt(None) == Config.ENABLE_TAG_DETAIL_DEFAULT
    config.enable_tag_detail(True)
    assert config.get_enable_tag_detail() == True
    assert Config.get_enable_tag_detail_alt(config) == True
    assert Config.get_enable_tag_detail_alt(None) == Config.ENABLE_TAG_DETAIL_DEFAULT

    assert config.get_enable_only_json() == Config.ENABLE_ONLY_JSON_DEFAULT
    assert Config.get_enable_only_json_alt(config) == Config.ENABLE_ONLY_JSON_DEFAULT
    assert Config.get_enable_only_json_alt(None) == Config.ENABLE_ONLY_JSON_DEFAULT
    config.enable_only_json()
    assert config.get_enable_only_json() == True
    assert Config.get_enable_only_json_alt(config) == True
    assert Config.get_enable_only_json_alt(None) == Config.ENABLE_ONLY_JSON_DEFAULT
    config.enable_only_json(False)
    assert config.get_enable_only_json() == False
    assert Config.get_enable_only_json_alt(config) == False
    assert Config.get_enable_only_json_alt(None) == Config.ENABLE_ONLY_JSON_DEFAULT
    config.enable_only_json(True)
    assert config.get_enable_only_json() == True
    assert Config.get_enable_only_json_alt(config) == True
    assert Config.get_enable_only_json_alt(None) == Config.ENABLE_ONLY_JSON_DEFAULT

    assert config.get_infinity() == Config.INFINITY_DEFAULT
    assert Config.get_infinity_alt(config) == Config.INFINITY_DEFAULT
    assert Config.get_infinity_alt(None) == Config.INFINITY_DEFAULT
    config.set_infinity('abc')
    assert config.get_infinity() == 'abc'
    assert Config.get_infinity_alt(config) == 'abc'
    assert Config.get_infinity_alt(None) == Config.INFINITY_DEFAULT
    config.set_infinity(None)
    assert config.get_infinity() is None
    assert Config.get_infinity_alt(config) is None
    assert Config.get_infinity_alt(None) == Config.INFINITY_DEFAULT

    assert config.get_nan() == Config.NAN_DEFAULT
    assert Config.get_nan_alt(config) == Config.NAN_DEFAULT
    assert Config.get_nan_alt(None) == Config.NAN_DEFAULT
    config.set_nan('Xyz')
    assert config.get_nan() == 'Xyz'
    assert Config.get_nan_alt(config) == 'Xyz'
    assert Config.get_nan_alt(None) == Config.NAN_DEFAULT
    config.set_nan(None)
    assert config.get_nan() is None
    assert Config.get_nan_alt(config) is None
    assert Config.get_nan_alt(None) == Config.NAN_DEFAULT

    assert config.get_max_array_size() == Config.MAX_ARRAY_SIZE_DEFAULT
    assert Config.get_max_array_size_alt(config) == Config.MAX_ARRAY_SIZE_DEFAULT
    assert Config.get_max_array_size_alt(None) == Config.MAX_ARRAY_SIZE_DEFAULT
    config.set_max_array_size(20)
    assert config.get_max_array_size() == 20
    assert Config.get_max_array_size_alt(config) == 20
    assert Config.get_max_array_size_alt(None) == Config.MAX_ARRAY_SIZE_DEFAULT

    assert config.get_max_depth() == Config.MAX_DEPTH_DEFAULT
    assert Config.get_max_depth_alt(config) == Config.MAX_DEPTH_DEFAULT
    assert Config.get_max_depth_alt(None) == Config.MAX_DEPTH_DEFAULT
    config.set_max_depth(8)
    assert config.get_max_depth() == 8
    assert Config.get_max_depth_alt(config) == 8
    assert Config.get_max_depth_alt(None) == Config.MAX_DEPTH_DEFAULT

    assert config.get_max_loop_times() == Config.MAX_LOOP_TIMES_DEFAULT
    assert Config.get_max_loop_times_alt(config) == Config.MAX_LOOP_TIMES_DEFAULT
    assert Config.get_max_loop_times_alt(None) == Config.MAX_LOOP_TIMES_DEFAULT
    config.set_max_loop_times(15)
    assert config.get_max_loop_times() == 15
    assert Config.get_max_loop_times_alt(config) == 15
    assert Config.get_max_loop_times_alt(None) == Config.MAX_LOOP_TIMES_DEFAULT



"""Statement for VivJson

Refer to:
- "Crafting Interpreters"
  https://craftinginterpreters.com/
Note that this code is made from scratch. The source code
of the above WEB site is not used.

Environment:
- Python 3.9 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-29'

from dataclasses import dataclass
from typing import Optional, TypeVar, Any
from .tokens import Token

@dataclass(frozen=True)
class Statement:
    """Statement data class

    Abstract class.
    """
    def __repr__(self):
        return ''

LITERAL = TypeVar("LITERAL", str, int, float, bool, None)

@dataclass(frozen=True)
class Literal(Statement):
    """Literal data class"""
    token: Token

    def __repr__(self):
        if self.token.type in (Token.Type.NULL,
                               Token.Type.TRUE, Token.Type.FALSE):
            return self.token.type.value
        mark = '"' if self.token.type == Token.Type.STRING else ''
        return f'{mark}{self.token.lexeme}{mark}'

@dataclass(frozen=True)
class Identifier(Statement):
    """Identifier data class

    It is used as name of variable, function, and class.
    """
    name: Token

    def __repr__(self):
        return self.name.lexeme

@dataclass(frozen=True)
class Keyword(Statement):
    """Keyword data class

    return, break, and so on
    """
    token: Token

    def __repr__(self):
        return self.token.lexeme

@dataclass(frozen=True)
class Blank(Statement):
    """Blank data class"""
    token: Token

    BLANK = '(blank)'

    def __repr__(self):
        return self.BLANK

@dataclass(frozen=True)
class Array(Statement):
    """Array data class"""
    values: list[Statement]

    def __repr__(self):
        return f'[{", ".join(map(str, self.values))}]'

@dataclass(frozen=True)
class Block(Statement):
    """Block data class

    For example,
    - anonymous: x = {a: 3, b: 2}
    - pure: function test() {return(10)}
    - limited: if (i > 10) {x="+"} else {x="-"}

    In VivJson, any block is function.
    The type of the outermost block (that is given from file/text) is
    decided as anonymous function or class constructor by calling the
    particular method.
    """
    values: list[Statement]
    type: str

    CLASS_CONSTRUCTOR = 'Block_class'
    ANONYMOUS_FUNCTION = 'Block_anonymous'
    PURE_FUNCTION = 'Block_pure'
    LIMITED_FUNCTION = 'Block_limited'

    def __repr__(self):
        return f'{{{", ".join(map(str, self.values))}}}'

@dataclass(frozen=True)
class Binary(Statement):
    """Binary data class

    For example, 3 * 2 are stored into left, operator, and right.
    """
    left: Statement
    operator: Token
    right: Statement

    def __repr__(self):
        return f'({self.left} {self.operator.lexeme} {self.right})' \
            if self.operator.type != Token.Type.NOT else f'(not {self.right})'

@dataclass(frozen=True)
class Parameter(Statement):
    """Parameter data class

    This is used to assist the definition of function entity.
    The example is shown as below. There are 3 Parameters.

    (1) modifier + name   (3) modifier + name
         |                     |
         |    (2) only name    |
         |          |          |
         V          V          v
    _____________ ____  ______________
    function test(data, function block) {
        :
        :
    }

    - "modifier" is essential at (1).
    - "modifier" is optional at (2) and (3).
    """
    modifier: Optional[Token]
    name: Identifier

    def __repr__(self):
        modifier = f'{self.modifier.lexeme} ' \
                    if self.modifier is not None else ''
        return modifier + self.name.name.lexeme

@dataclass(frozen=True)
class Callee(Statement):
    """Callee data class

    This is used to define function entity.
    The example is shown as below.

        name   parameters[0]   parameters[1]
         |         |             |
         V         V             v
    _____________ ____  ______________
    function test(data, function block) {
        :
        :
    }

    In "name", its "modifier" of Parameter is essential.
    In "parameters", its "modifier" of Parameter is optional.
    """
    name: Parameter
    parameters: Array

    def __repr__(self):
        return f'{self.name}({", ".join(map(str, self.parameters.values))})'

@dataclass(frozen=True)
class CalleeRegistry(Statement):
    """Callee Registry data class

    This is made in Evaluator. This is not used in Parser.

    Member "environment" is Environment instance or None.
    The former (Environment instance) is set if callee is Closure.

    Member "is_reference" is True if this is not the definition of
    function.

    For example, "enclosure", "z1", and "z2" are registered as this
    data class.
    "environment" of "enclosure" is None.
    On the other hand, "environment" of "z1" and "z2" is Environment
    instance.
    "is_reference" of "enclosure" is False. "is_reference" of "z1" and
    "z2" is True.

    function enclosure(a) {
        x = a
        function closure(y) {
            return(x + y)
        }
        return(closure)
    }
    z1 = enclosure(100)
    z2 = enclosure(200)
    print(z1(5))  # 105
    print(z2(10))  # 210

    By the way, this data class is used even if assignment is simple.
    The following "y" is also this instance.
    "environment" of "x2" and "y" is None.
    "is_reference" of "x2" is False. "is_reference" of "y" is True.

    function x2(a) {
        return(a * 2)
    }
    y = x2

    "is_reference" decides whether the variable is remained or not
    after evaluating block.
    In the following sample, member of "k" will be only "y" after
    evaluation. Because the definition of function is not remained
    as result. Although value of "x2" and "y" is this instance,
    since "is_reference" of "y" is True, only "y" will be remained.

    k = {
        function x2(a) {
            return(a * 2)
        }
        y = x2
    }

    Similarly, the following "table" can keep its members.

    function add(a, b) {
        return(a + b)
    }
    function sub(a, b) {
        return(a - b)
    }
    table = {
        "+": add,
        "-": sub
    }
    """
    callee: Callee
    environment: Any
    is_reference: bool

    def __repr__(self):
        return str(self.callee)

@dataclass(frozen=True)
class Call(Statement):
    """Call data class

    This is used to call function, such as len("abc").
    """
    name: Statement
    arguments: Array

    def __repr__(self):
        return f'{self.name}({", ".join(map(str, self.arguments.values))})'

@dataclass(frozen=True)
class Loop(Statement):
    """Loop data class

    "call" is not used for actual control. It may be used to print
    function name.
    "statements" is main operation.
    "initial" proposes whether loop should be done or not before loop.
    "continuous" proposes whether loop should be done or not after 1st
    loop.
    "initial" and "continuous" are a list of Statement. The judgement
    is decided with the last Statement.
    "each" and "iterator" is used for "for (i in [1, 2, 3]) {...}"
    style. "each" is variable as Identifier. "iterator" is list or
    dict.

    For example, for (i = 0; i < 10; i += 1) { print(i) }
    "initial" is [i = 0; i < 10].
    "continuous" is [i += 1; i < 10].
    "statements" is [print(i)].
    "each" is None.
    "iterator" is None.

    For example, for (i in [1, 2, 3]) { print(i) }
    "initial" is [true].
    "continuous" is [true].
    "statements" is [print(i)].
    "each" is i.
    "iterator" is [1, 2, 3].

    For example, for (i in {"a": 1, "b": 2, "c": 3}) { print(i) }
    "initial" is [true].
    "continuous" is [true].
    "statements" is [print(i)].
    "each" is i.
    "iterator" is {"a": 1, "b": 2, "c": 3}.
    """
    call: Call
    initial: list[Statement]
    continuous: list[Statement]
    statements: list[Statement]
    each: Optional[Identifier]
    iterator: Optional[list]

    def __repr__(self):
        if isinstance(self.each, Identifier) and self.iterator is not None:
            text = ", ".join(map(str, self.statements))
            return f'{self.call.name}({self.each} in {self.iterator}) ' \
                   f'{{{text}}}'
        texts = []
        for statements in [self.initial, self.continuous, self.statements]:
            text = ", ".join(map(str, statements))
            texts.append(f'{{{text}}}')
        return f'{self.call.name}(init={texts[0]}, continue={texts[1]}) ' \
               f'{texts[2]}'

@dataclass(frozen=True)
class Get(Statement):
    """Get data class

    For example, x["y"]["z"] is represented as ["x", "y", "z"].
    Similarly, x.y.z is represented.
    """
    members: list[Statement]

    def __repr__(self):
        variable = ''
        for member in self.members:
            variable = str(member) if variable == '' \
                        else f'{variable}[{member}]'
        return variable

@dataclass(frozen=True)
class Set(Statement):
    """Set data class

    For example, x["y"]["z"] = 3 is represented as below.
      - members: ["x", "y", "z"]
      - operator: =
      - value: 3
    Similarly, x.y.z = 3 is represented.
    """
    members: list[Statement]
    operator: Token
    value: Statement

    def __repr__(self):
        variable = ''
        for member in self.members:
            variable = str(member) if variable == '' \
                        else f'{variable}[{member}]'
        space = '' if self.operator.type == Token.Type.SET else ' '
        return f'{variable}{space}{self.operator.lexeme} {self.value}'

@dataclass(frozen=True)
class Remove(Statement):
    """Remove data class

    For example, remove(x["y"]["z"]), remove(x.y.z)
    """
    token: Token
    members: list[Statement]

    def __repr__(self):
        variable = ''
        for member in self.members:
            variable = str(member) if variable == '' \
                        else f'{variable}[{member}]'
        return f'remove({variable})'

@dataclass(frozen=True)
class Return(Statement):
    """Return data class

    "return" [ "(" value ")" ]
    """
    token: Token
    value: Optional[Statement]

    def __repr__(self):
        if self.value is None:
            return 'return'
        value = str(self.value)
        if value[0] == '(' and value[-1] == ')':
            value = value[1:-1]
        return f'return({value})'

# tuple is disallowed.
# When tuple has simple structure, such as 1 dimensional tuple,
# converting to list is possible. However complex structure is
# given, it is so difficult. So it is not allowed.
HOSTVALUE = \
    TypeVar("HOSTVALUE", None, bool, int, float, str, list, dict)

@dataclass(frozen=True)
class Injection(Statement):
    """Injection data class

    Set host language's variable.
    """
    variable: str
    value: HOSTVALUE
    location: Optional[str]  # for reporting Error, ex) "2nd argument"

    def __repr__(self):
        value = self.value
        if value is None:
            value = 'null'
        elif value is True:
            value = 'true'
        elif value is False:
            value = 'false'
        return f'{self.variable}: {value}'

@dataclass(frozen=True)
class Value(Statement):
    """Value data class

    Set host language's value.
    """
    value: HOSTVALUE
    location: Optional[str]  # for reporting Error, ex) "2nd argument"

    def __repr__(self):
        value = self.value
        if value is None:
            value = 'null'
        elif value is True:
            value = 'true'
        elif value is False:
            value = 'false'
        return str(value)



"""Command-line client for VivJson

Usage: vivjson [ [option] <code or file> ... ]
         or
       python3 -m vivjson [ [option] <code or file> ... ]

option: - "+" is concatenated the previous code and
          the following code, such as
            "a=" + "[1, 2]" "return(a[0]+a[1])"
        - "-i" or "--stdin" reads from PIPE.
          For example, the following command shows "6".
            echo '{"a":3}' | vivjson -i "return(a*2)"
          "-i=<name>" or "--stdin=<name>" gives variable name for
          value that read from PIPE.
          For example, the following command shows
          {"dog": 2, "cat": 3}.
            echo '[{"x":"dog","y":2},{"x":"cat","y":3}]' | \
            vivjson -i=data "z={},for(v in data){z[v.x]=v.y},print(z)"
        - "-j" or "--json" indicates that the following code is JSON.
          For example, error occurs because the following value is
          invalid as JSON.
            vivjson -j '{"a":1+2}'
        - "-v" or "--version" shows version.

Environment:
- Python 3.9 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-28'

import sys
from .viv import Viv, Json
from .config import Config

def show_usage():
    """Show usage of VivJson in command line."""
    print(
        'VivJson\n'
        'Usage: vivjson [ [option] [ <code or file> ] ] ...\n'
        '         or\n'
        '       python3 -m vivjson [ [option] [ <code or file> ] ] ...\n'
        '\n'
        'option: - "+" is concatenated the previous code and the following\n'
        '          code, such as\n'
        '            "a=" + "[1, 2]" "return(a[0]+a[1])"\n'
        '        - "-i" or "--stdin" reads from PIPE.\n'
        '          For example, the following command shows "6".\n'
        '            echo \'{"a":3}\' | vivjson -i "return(a*2)"\n'
        '          "-i=<name>" or "--stdin=<name>" gives variable name for\n'
        '          value that read from PIPE.\n'
        '          For example, the following command shows\n'
        '          {"dog": 2, "cat": 3}.\n'
        '            echo \'[{"x":"dog","y":2},{"x":"cat","y":3}]\' | \\\n'
        '            vivjson -i=data "z={},for(v in data){z[v.x]=v.y},print(z)"\n'
        '        - "-j" or "--json" indicates that the following code is JSON.\n'
        '          For example, error occurs because the following value is\n'
        '          invalid as JSON.\n'
        '            vivjson -j \'{"a":1+2}\'\n'
        '        - "-v" or "--version" shows version.\n'
        '\n'
        '       Note that the file extension must be '
                  f'"{Viv.EXTENSIONS[0]}" or "{Viv.EXTENSIONS[1]}".\n'
        '\n'
        'Example 1. The following codes show same result (5).\n'
        '           - vivjson "a:3,b:2,return(a+b)"\n'
        '           - vivjson "{a:3,b:2,return(a+b)}"\n'
        '           - vivjson "{a:3,b:2}" "{return(a+b)}"\n'
        '           - vivjson "{a:3,b:2}" "return(a+b)"\n'
        '           - vivjson "x=" + "{a:3,b:2}" "return(x.a+x.b)"\n'
        '           - vivjson "3,2" "return(_[0]+_[1])"\n'
        '           - vivjson 3 2 "return(_[0]+_[1])"\n'
        '           - vivjson 3 "return(_ + 2)"\n'
        'Example 2. Using file.\n'
        '           - vivjson test.viv\n'
        '           - vivjson data.json calc.viv\n'
        'Example 3. Using code and file.\n'
        '           - vivjson "{a:3,b:2}" calc.viv\n'
        'Example 4. Using PIPE.\n'
        '           - echo "return(3*5)" | vivjson -i\n'
        '           - echo "a=3" | vivjson -i "return(a*2)"\n'
        '           - echo \'{"a":3}\' | vivjson -i "return(a*2)"\n'
        '           - cat test.viv | vivjson -i\n'
        '           - cat data.json | vivjson -i "return(a*b)"\n'
        'Example 5. Parsing data as JSON.\n'
        '           - vivjson -j \'{"a":3,"b":2}\' "return(a+b)"\n'
        '           - echo \'{"a":3,"b":2}\' | vivjson -j -i "return(a+b)"\n'
        '\n'
        '       Note that the combined option, such as "-ji", isn\'t allowed.'
    )

def show_version():
    """Show version of VivJson in command line."""
    print(
        "VivJson's\n"
        f'    specification version: {Config.SPEC_VERSION}\n'
        f'    interpreter version:   {Config.INTERPRETER_VERSION}'
    )

def get_stdin():
    """Get data or command from stdin.

    Returns:
        str or NoneType: data/command if it is given,
                         None otherwise.
    """
    data = []
    try:
        while True:
            line = input().strip()
            data.append(line)
    except EOFError:
        pass

    if len(data) == 0:
        return None
    return '\n'.join(data)

def is_number(text: str) -> bool:
    """Detect whether the given text is number or not.

    Args:
        text (str): A text that may be number

    Returns:
        bool: True if the given text is number, False otherwise
    """
    try:
        float(text)
    except (ValueError, OverflowError):
        return False
    return True

def main():  # pylint: disable=R0912
    """Main function for command line."""
    if len(sys.argv) < 2:
        show_usage()
        return

    arguments = sys.argv.copy()
    arguments.pop(0)

    for argument in arguments:
        if argument in ('-v', '--version'):
            show_version()
            return

    stdin_index = -1
    stdin_name = None
    for index, argument in enumerate(arguments):
        for option in ('-i', '--stdin'):
            if argument.find(option) == 0:
                stdin_index = index
                if '=' in argument:
                    data = argument.split('=')
                    if len(data) == 2 and len(data[1]) > 0:
                        stdin_name = data[1]
                break
    if stdin_index >= 0:
        arguments.pop(stdin_index)

        stdin = get_stdin()
        if not isinstance(stdin, str):
            error_message = Viv.report_error('Cannot read from stdin.')
            print(error_message, file=sys.stderr)
            return
        if stdin_name is not None:
            stdin = f'{stdin_name}={stdin}'
        arguments.insert(0, stdin)  # #1

    # | Original | Stdin | arguments     | stdinIndex | array as Result |
    # |----------|-------|---------------|------------|-----------------|
    # | "a:3"    | ----  | ["a:3"]       | -1         | ["a:3"]         |
    # | -j "a:3" | ----  | ["-j", "a:3"] | -1         | [Json("a:3")]   |
    # | -i       | "a:3" | ["a:3"]       |  0         | ["a:3"]         |
    # | -j -i    | "a:3" | ["a:3", "-j"] |  1         | [Json("a:3")]   |
    array = []
    index = -1
    while (index := index + 1) < len(arguments):
        if len(argument) >= 1 and argument[0] == '#':
            break

        argument = arguments[index]
        if argument in ('-j', '--json'):
            if index == stdin_index:
                if index == 0:
                    error_message = Viv.report_error('Unexpected behavior.')
                    print(error_message, file=sys.stderr)
                    return
                array[0] = Json(arguments[0])  # #1
                continue
            if index + 1 < len(arguments):
                index += 1
                array.append(Json(arguments[index]))
                continue
        array.append(argument)

    if len(array) == 1 and isinstance(array[0], str) \
            and (len(array[0]) == 0
                 or (array[0][0] == '-' and not is_number(array[0]))):
        show_usage()
        return

    result, error_message = Viv.run(array)
    if error_message != '':
        print(error_message, file=sys.stderr)
        return

    print(Viv.make_string(result))

if __name__ == '__main__':
    main()



"""Configuration for VivJson

Environment:
- Python 3.9 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-25'

from typing import Optional

class Config:  # pylint: disable=R0902, R0904
    """Config class

    Attributes:
        _enable_stderr (bool): When True is given, error message is
                               outputted into stderr.
                               Otherwise, it is not outputted.
        _enable_tag_detail (bool): When True is given, error message's
                               tag contains either of "Lexer",
                               "Parser", or "Evaluator".
        _enable_only_json (bool): When true is given, the given data is
                                  parsed as JSON.
                                  In other words, script is disabled.
        _infinity (str on NoneType): When string is given, Infinity is
                                     allowed. Then the given string is
                                     used to input/output Infinity
                                     from/to JSON. (Note that it is not
                                     surrounded with quotation mark.)
                                     When None is given and Infinity is
                                     happen, error is occurred.
        _nan (str on NoneType): When string is given, NaN (Not a
                                Number) is allowed. Then the given
                                string is used to input/output NaN
                                from/to JSON. (Note that it is not
                                surrounded with quotation mark.)
                                When None is given and NaN is happen,
                                error is occurred.
        _max_array_size (int): Maximum array/block size
        _max_depth (int): Maximum recursive called times of evaluate
                          method
        _max_loop_times (int): Maximum loop times of "for", "while",
                               and so on
        SPEC_VERSION (str): Version of VivJson's specification
                            as <Major> "." <Minor> "." <Patch>
                            (For example, "1.0.0")
        INTERPRETER_VERSION (str): Version of this interpreter as
                                   <Major> "." <Minor> "." <Patch>
                                   (For example, "1.0.0")
                                   This <Major> "." <Minor> is equal to
                                   <SPEC_VERSION>'s one.
        ENABLE_STDERR_DEFAULT (bool): The default of enabling error
                                      message's stderr output
        ENABLE_TAG_DETAIL_DEFAULT (bool): The default of enabling
                                      detail of error message's tag
        ENABLE_ONLY_JSON_DEFAULT (bool): The default of enabling only
                                         JSON
        INFINITY_DEFAULT (str on NoneType): The default of Infinity
        NAN_DEFAULT (str on NoneType): The default of NaN (Not a
                                       Number)
        MAX_ARRAY_SIZE_DEFAULT (int): The default of maximum array size
        MAX_DEPTH_DEFAULT (int): The default of maximum recursive
                                 called times of evaluate method
        MAX_LOOP_TIMES_DEFAULT (int): The default of maximum loop times
                                      of "for", "while", and so on
    """
    SPEC_VERSION = '1.0.0'
    INTERPRETER_VERSION = '1.0.0'
    ENABLE_STDERR_DEFAULT = False
    ENABLE_TAG_DETAIL_DEFAULT = False
    ENABLE_ONLY_JSON_DEFAULT = False
    INFINITY_DEFAULT = None
    NAN_DEFAULT = None
    MAX_ARRAY_SIZE_DEFAULT = 1000
    MAX_DEPTH_DEFAULT = 200
    MAX_LOOP_TIMES_DEFAULT = 1000

    def __init__(self,  # pylint: disable=R0913
                 enable_stderr=ENABLE_STDERR_DEFAULT,
                 enable_tag_detail=ENABLE_TAG_DETAIL_DEFAULT,
                 enable_only_json=ENABLE_ONLY_JSON_DEFAULT,
                 infinity=INFINITY_DEFAULT,
                 nan=NAN_DEFAULT,
                 max_array_size=MAX_ARRAY_SIZE_DEFAULT,
                 max_depth=MAX_DEPTH_DEFAULT,
                 max_loop_times=MAX_LOOP_TIMES_DEFAULT):
        """Initialize class.

        Args:
            enable_stderr (bool, optional): When True is given, error 
                                    message is outputted into stderr.
                                    Otherwise, it is not outputted.
                                    The default is False.
            enable_tag_detail (bool): When True is given, error
                                    message's tag contains either of
                                    "Lexer", "Parser", or "Evaluator".
            enable_only_json (bool): When true is given, the given data
                                     is parsed as JSON.
                                     In other words, script is
                                     disabled.
            infinity (str on NoneType, optional): When string is given,
                                    Infinity is allowed in JSON. Then
                                    the given string is used to
                                    input/output Infinity from/to JSON.
                                    (Note that it is not surrounded
                                    with quotation mark.)
                                    When None is given and Infinity is
                                    happen, error may be occurred.
                                    The default is None.
            nan (str on NoneType, optional): When string is given,
                                    NaN (Not a Number) is allowed in
                                    JSON. Then the given string is used
                                    to input/output NaN from/to JSON.
                                    (Note that it is not surrounded
                                    with quotation mark.)
                                    When None is given and NaN is
                                    happen, error may be occurred.
                                    The default is False.
            max_array_size (int, optional): Maximum array/block size
            max_depth (int, optional): Maximum recursive called times
                                       of evaluate method
            max_loop_times (int, optional): Maximum loop times of
                                            "for", "while", and so on
        """
        self._enable_stderr = enable_stderr
        self._enable_tag_detail = enable_tag_detail
        self._enable_only_json = enable_only_json
        self._infinity = infinity
        self._nan = nan
        self._max_array_size = max_array_size
        self._max_depth = max_depth
        self._max_loop_times = max_loop_times

    def clone(self):
        """Copy itself.

        Returns:
            Config: Copied Config object
        """
        return Config(self._enable_stderr,
                      self._enable_tag_detail,
                      self._enable_only_json,
                      self._infinity,
                      self._nan,
                      self._max_array_size,
                      self._max_depth,
                      self._max_loop_times)

    def enable_stderr(self, enable=True):
        """Enable that error message outputs into stderr.

        Args:
            enable (bool, optional): When True is given, error message
                                     is outputted into stderr.
                                     Otherwise, it is not outputted.
                                     The argument's default is True.
        """
        self._enable_stderr = enable

    def get_enable_stderr(self) -> bool:
        """Get whether outputting error message into stderr is enable.
        """
        return self._enable_stderr

    @staticmethod
    def get_enable_stderr_alt(config) -> bool:
        """Get whether outputting error message into stderr is enable.

        Args:
            config (Config or NoneType): Instance of configuration.
                                         When it is None, default
                                         value is returned.

        Returns:
            bool: Outputting error message into stderr is enable or not
        """
        return Config.ENABLE_STDERR_DEFAULT if config is None \
               else config.get_enable_stderr()

    def enable_tag_detail(self, enable=True):
        """Enable that detail of error message's tag.

        Args:
            enable (bool, optional): When True is given, error
                                     message's tag contains either of
                                     "Lexer", "Parser", or "Evaluator".
                                     The argument's default is True.
        """
        self._enable_tag_detail = enable

    def get_enable_tag_detail(self) -> bool:
        """Get whether detail of error message's tag is enable or not.
        """
        return self._enable_tag_detail

    @staticmethod
    def get_enable_tag_detail_alt(config) -> bool:
        """Get whether detail of error message's tag is enable or not.

        Args:
            config (Config or NoneType): Instance of configuration.
                                         When it is None, default
                                         value is returned.

        Returns:
            bool: Detail of error message's tag is enable or not
        """
        return Config.ENABLE_TAG_DETAIL_DEFAULT if config is None \
               else config.get_enable_tag_detail()

    def enable_only_json(self, enable=True):
        """Enable that parsing data as only JSON.

        Args:
            enable (bool, optional): When true is given, the given data
                                     is parsed as JSON.
                                     In other words, script is
                                     disabled.
        """
        self._enable_only_json = enable

    def get_enable_only_json(self) -> bool:
        """Get whether parsing data as only JSON is enable or not.
        """
        return self._enable_only_json

    @staticmethod
    def get_enable_only_json_alt(config) -> bool:
        """Get whether parsing data as only JSON is enable or not.

        Args:
            config (Config or NoneType): Instance of configuration.
                                         When it is None, default
                                         value is returned.

        Returns:
            bool: parsing data as only JSON is enable or not
        """
        return Config.ENABLE_ONLY_JSON_DEFAULT if config is None \
               else config.get_enable_only_json()

    def set_infinity(self, infinity: Optional[str]):
        """Set infinity's string.

        When string is given, Infinity is allowed. Then the given
        string is used to input/output Infinity from/to JSON. (Note
        that it is not  surrounded with quotation mark.)
        When None is given and Infinity is happen, error may be
        occurred.

        Args:
            infinity (str or NoneType): Infinity's string or None.
        """
        self._infinity = infinity

    def get_infinity(self) -> Optional[str]:
        """Get infinity's string."""
        return self._infinity

    @staticmethod
    def get_infinity_alt(config) -> Optional[str]:
        """Get infinity's string.

        Args:
            config (Config or NoneType): Instance of configuration.
                                         When it is None, default
                                         value is returned.

        Returns:
            str on NoneType: Infinity's string.
                             When this is None, Infinity is not
                             allowed.
        """
        return Config.INFINITY_DEFAULT if config is None \
               else config.get_infinity()

    def set_nan(self, nan: Optional[str]):
        """Set NaN(Not a Number)'s string.

        When string is given, NaN (Not a Number) is allowed. Then the
        given string is used to input/output NaN from/to JSON. (Note
        that it is not surrounded with quotation mark.)
        When None is given and NaN is happen, error may be occurred.

        Args:
            nan (str or NoneType): NaN(Not a Number)'s string
        """
        self._nan = nan

    def get_nan(self) -> Optional[str]:
        """Get NaN(Not a Number)'s string."""
        return self._nan

    @staticmethod
    def get_nan_alt(config) -> Optional[str]:
        """Get NaN(Not a Number)'s string.

        Args:
            config (Config or NoneType): Instance of configuration.
                                         When it is None, default
                                         value is returned.

        Returns:
            str on NoneType: NaN (Not a Number)'s string.
                             When this is None, NaN is not allowed.
        """
        return Config.NAN_DEFAULT if config is None else config.get_nan()

    def set_max_array_size(self, size: int):
        """Set maximum array/block size."""
        self._max_array_size = size

    def get_max_array_size(self) -> int:
        """Get maximum array/block size."""
        return self._max_array_size

    @staticmethod
    def get_max_array_size_alt(config) -> int:
        """Get maximum array/block size.

        Args:
            config (Config or NoneType): Instance of configuration.
                                         When it is None, default
                                         value is returned.

        Returns:
            int: Maximum array/block size
        """
        return Config.MAX_ARRAY_SIZE_DEFAULT if config is None \
               else config.get_max_array_size()

    def set_max_depth(self, depth: int):
        """Set maximum recursive called times of evaluate method."""
        self._max_depth = depth

    def get_max_depth(self) -> int:
        """Get maximum recursive called times of evaluate method."""
        return self._max_depth

    @staticmethod
    def get_max_depth_alt(config) -> int:
        """Get maximum recursive called times of evaluate method.

        Args:
            config (Config or NoneType): Instance of configuration.
                                         When it is None, default
                                         value is returned.

        Returns:
            int: Maximum recursive called times of evaluate method
        """
        return Config.MAX_DEPTH_DEFAULT if config is None \
               else config.get_max_depth()

    def set_max_loop_times(self, times: int):
        """Set maximum loop times of "for", "while", and so on."""
        self._max_loop_times = times

    def get_max_loop_times(self) -> int:
        """Get maximum loop times of "for", "while", and so on."""
        return self._max_loop_times

    @staticmethod
    def get_max_loop_times_alt(config) -> int:
        """Get maximum loop times of "for", "while", and so on.

        Args:
            config (Config or NoneType): Instance of configuration.
                                         When it is None, default
                                         value is returned.

        Returns:
            int: Maximum loop times of "for", "while", and so on.
        """
        return Config.MAX_LOOP_TIMES_DEFAULT if config is None \
               else config.get_max_loop_times()



"""API of VivJson

- Viv.run: Run VivJson's code or deserialize JSON objects.
- Viv.parse: Parse VivJson's code and JSON object.
- Viv.parse_file: Parse a file that contains VivJson's code or JSON object.
- Viv.parse_text: Parse a text that is VivJson's code or JSON object.
- Viv.make_instance: Make a class instance.
- Viv.make_string: Convert into String. Serialize into JSON string.

Environment:
- Python 3.9 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-30'

# pylint: disable=W0718, W4901, C0411

from dataclasses import dataclass
import os
from typing import Optional, TypeVar
from .config import Config
from .environment import Environment
from .error import Error
from .evaluator import Evaluator
from .parser import Parser
from .statement import Array, Blank, Block, Call, CalleeRegistry, \
                       Identifier, Injection, Literal, Return, \
                       Set, Statement, Value
from .tokens import Token

class Viv:
    """Viv class

    Attributes:
        EXTENSIONS (tuple[str]): The effective file extensions.
    """
    EXTENSIONS = ('.viv', '.json')

    @staticmethod
    def run(*parameters):
        """Run VivJson's code or deserialize JSON objects.

        For example,
        - run("a:3,b:2,return(a+b)")
        - run(["{a:3,b:2}", "{return(a+b)}"])
        - run(["x=", "+", "{a:3,b:2}", "return(x.a+x.b)"])
        - run("test.viv")
        - run(["test.viv"])
        - run("data.json", "calc.viv")
        - run(["data.json", "calc.viv"])
        - run("{a:3,b:2}", "calc.viv")
        - run("x/", Config(enable_stderr=True))  # Error at run-time
        - run({"x":[1,2],"y":True},"return(if(y){:=x[0]}else{:=x[1]})")

        Args:
            parameters (Any): VivJson's codes, JSON values,
                              file paths, variables,
                              Parsed objects, a class instance,
                              a method object, or a Configuration.
                            str: A VivJson's code, a JSON value, or
                                 a file path
                            list[Any] or tuple[Any]: Some VivJson's
                                                codes, JSON values,
                                                some file paths,
                                                variables, or
                                                Parsed objects
                            dict: Some variables that are name/value
                                  pairs.
                            Json: JSON object
                            Parsed: The parsed object that is generated
                                    with Viv.parse, Viv.parse_text, or
                                    Vir.parse_file method.
                            Instance: Class instance. It is passed
                                      through toward output.
                            Method: Calling class method.
                                    This object has 2 members.
                                    "name" is class method name.
                                    "arguments" is its arguments.
                            Config: Configuration

        Returns:
            Any, str: Result and Error message.
                      1st value: Result of the given codes if success,
                                 None otherwise
                      2nd value: "" if success,
                                 Error message otherwise.
        """
        # Multiple parameters are represented as tuple.
        # | Caller of this method   | parameters            |
        # |-------------------------|-----------------------|
        # | run("x = 3")            | ("x = 3",)            |
        # | run("x = 3", "y = 5")   | ("x = 3", "y = 5")    |
        # | run(["x = 3", "y = 5"]) | (["x = 3", "y = 5"],) |
        parsed, config, error_message = Viv._parse_internal(parameters)
        if error_message != '':
            return None, error_message

        # Evaluate.
        value, error_message = Viv._run_internal(parsed, config)
        if error_message != '':
            return None, error_message
        value = Viv._collect_only_valid_elements(value)
        if value == Environment.UNDEFINED:
            return None, Viv.report_error('The returned value is invalid')
        return value, ''

    @staticmethod
    def parse(*parameters):
        """Parse VivJson's code and JSON object.

        For example,
        - parse("a:3,b:2,return(a+b)")
        - parse(["{a:3,b:2}", "{return(a+b)}"])
        - parse(["x=", "+", "{a:3,b:2}", "return(x.a+x.b)"])
        - parse("test.viv")
        - parse(["test.viv"])
        - parse("data.json", "calc.viv")
        - parse(["data.json", "calc.viv"])
        - parse("{a:3,b:2}", "calc.viv")
        - parse("x/", Config(enable_stderr=True))  # Error at run-time
        - parse({"x":[1,2],"y":True},"return(if(y){:=x[0]}else{:=x[1]})")

        Args:
            parameters (Any): VivJson's codes, JSON values,
                              file paths, variables,
                              Parsed objects, a class instance,
                              a method object, or a Configuration.
                            str: A VivJson's code, a JSON value, or
                                 a file path
                            list[Any] or tuple[Any]: Some VivJson's
                                                codes, JSON values,
                                                some file paths,
                                                variables, or
                                                Parsed objects
                            Json: JSON object
                            dict: Some variables that are name/value
                                  pairs.
                            Parsed: The parsed object that is generated
                                    with this method. It is passed
                                    through toward output.
                            Instance: Class instance. It is passed
                                      through toward output.
                            Method: Class method information. It is
                                    passed through toward output.
                            Config: Configuration

        Returns:
            Parsed or NoneType, str: Parsed object of script and
                                     Error message.
                                   1st value: Parsed object if success,
                                              None otherwise
                                              Parsed object has
                                              statements of the given
                                              codes 
                                   2nd value: "" if success,
                                              Error message otherwise.
        """
        # Multiple parameters are represented as tuple.
        # | Caller of this method     | parameters            |
        # |---------------------------|-----------------------|
        # | parse("x = 3")            | ("x = 3",)            |
        # | parse("x = 3", "y = 5")   | ("x = 3", "y = 5")    |
        # | parse(["x = 3", "y = 5"]) | (["x = 3", "y = 5"],) |
        parsed, _, error_message = Viv._parse_internal(parameters)
        return parsed, error_message

    @staticmethod
    def _parse_internal(parameters):  # pylint: disable=R0912, R0914, R0915
        """Parse VivJson's code and JSON object.

        Args:
            parameters (list or tuple): VivJson's codes, JSON values,
                                        file paths, variables, Parsed
                                        objects, a class instance,
                                        a method object,
                                        or a Configuration.

        Returns:
            Parsed or NoneType, Config or NoneType, str:
                        Parsed object of script, Configuration, and
                        Error message.
                        1st value: Parsed object if paring is success,
                                   None otherwise.
                                   Parsed object has statements of
                                   the given codes 
                        2nd value: Configuration if it is given,
                                   None otherwise
                        3rd value: "" if success,
                                   Error message otherwise.
        """
        extracted_parameters, instance, config = \
                                        Viv._extract_parameters(parameters)
        if extracted_parameters is None:
            return None, config, Viv.report_error('Invalid parameter')

        # Concatenate if + option is given.
        targets, error_message = \
                        Viv._concatenate_parameters(extracted_parameters)
        if error_message != '':
            return None, config, error_message

        # Parse the given target.
        statements_array = []
        for target in targets:
            # target[0]: index of the given argument
            # target[1]: parameter
            statements = None
            error_message = ''
            if isinstance(target[1], dict):
                location = Viv._get_argument_text(target[0])
                statements = []
                for name, value in target[1].items():
                    if not isinstance(name, str) \
                            or not Viv._is_valid_type(value):
                        return None, config, Viv.report_error(
                                    f'Invalid parameter "{name}: {value}"')
                    statements.append(Injection(name, value, location))
            elif isinstance(target[1], Parsed):
                statements = target[1].statements
            elif Viv._get_file_extension(target[1]) is not None:
                parsed, error_message = Viv.parse_file(target[1], config)
                if parsed is not None:
                    statements = parsed.statements
            else:
                argument_index = None if len(targets) == 1 else target[0]
                text = None
                config_ext = None
                if isinstance(target[1], Json):
                    text = target[1].value
                    config_ext = Config() if config is None \
                                 else config.clone()
                    config_ext.enable_only_json()
                else:
                    text = target[1]
                    config_ext = config
                parsed, error_message = \
                        Viv.parse_text(text, config_ext, argument_index)
                if parsed is not None:
                    statements = parsed.statements
            if error_message != '':
                return None, config, error_message
            if not isinstance(statements, list):
                return None, config, Viv.report_error('No statement')
            statements_array.append(statements)

        # Collect implicit variables and fix them.
        implicit_variables = []
        for index, statements in enumerate(statements_array):
            if Viv._is_setting_implicit_variable(statements):
                implicit_variables.append(index)
        count = len(implicit_variables)
        if count >= 2:
            new_index = 0
            for index, statements in enumerate(statements_array):
                if index in implicit_variables:
                    Viv._fix_setting_implicit_variable(statements, new_index)
                    new_index += 1

            # Insert "_ = [null, ...]" before each assignment,
            # such as "_[0] = 1".
            identifier = Identifier(Token(Token.Type.IDENTIFIER, lexeme='_'))
            value = Array([Literal(Token(Token.Type.NULL))] * count)
            statements = [Set([identifier], Token(Token.Type.SET), value)]
            statements_array.insert(0, statements)

        # Concatenate statements.
        whole_statements = []
        for statements in statements_array:
            whole_statements.extend(statements)

        # Viv.print_statements(whole_statements)

        return Parsed(whole_statements, instance), config, ''

    @staticmethod
    def parse_file(file_path: str, config=None):
        """Parse a file.

        JSON value and VivJson's code can be parsed.

        Args:
            file_path (str): The path of file that has JSON value or
                             VivJson's code
            config (Config or NoneType, optional): Configuration if
                                                   needed,
                                                   None otherwise.

        Returns:
            Parsed or NoneType, str: Parsed object of script and
                                     Error message.
                                   1st value: Parsed object if success,
                                              None otherwise.
                                              Parsed object has
                                              statements of the given
                                              codes 
                                   2nd value: "" if success,
                                              Error message otherwise.
        """
        extension = Viv._get_file_extension(file_path)
        if extension is None:
            return None, Viv.report_error(f'"{file_path}" is invalid.')
        file_name = os.path.basename(file_path)

        code = None
        with open(file_path, encoding='utf-8') as file:
            code = file.read()

        if code is None:
            return None, Viv.report_error(
                                    f'Cannot read code from "{file_name}".')

        if extension == '.json':
            config_ext = Config() if config is None else config.clone()
            config_ext.enable_only_json()
            config = config_ext

        parser = Parser(code, file_name, config)
        error_message = ''
        try:
            statements = parser.parse()
            return Parsed(statements, None), ''
        except (Error.LexError, Error.ParseError) as e:
            error_message = str(e)
        except Exception as e:
            error_message = Viv.report_error(str(e))
        return None, error_message

    @staticmethod
    def parse_text(text: str, config=None, argument_index=None):
        """Parse a text.

        JSON value and VivJson's code can be parsed.

        Args:
            text (str): Text that is JSON value or VivJson's code
            config (Config or NoneType, optional): Configuration if
                                                   needed,
                                                   None otherwise.
            argument_index (int, optional): Argument index (0~)

        Returns:
            Parsed or NoneType, str: Parsed object of script and
                                     Error message.
                                   1st value: Parsed object if success,
                                              None otherwise
                                              Parsed object has
                                              statements of the given
                                              codes 
                                   2nd value: "" if success,
                                              Error message otherwise.
        """
        if not isinstance(text, str):
            return None, Viv.report_error(f'"{text}" is not string.')

        medium = None if argument_index is None \
                    else Viv._get_argument_text(argument_index)
        parser = Parser(text, medium, config)
        error_message = ''
        try:
            statements = parser.parse()
            return Parsed(statements, None), ''
        except (Error.LexError, Error.ParseError) as e:
            error_message = str(e)
        except Exception as e:
            error_message = Viv.report_error(str(e))
        return None, error_message

    @staticmethod
    def make_instance(*parameters):
        """Make a class instance.

        This method runs the given parameters as Constructor.
        Then its result will be class instance.

        For example, a class method is called as below.

        code = "function test(x, y) {" \
               "  z = x.a + x.b.1 + y" \
               "  return(z)" \
               "}"
        map_x = {"a": 100, "b": [1.0, 2.0]}
        instance, error_message = Viv.make_instance(code)
        result, error_message = \
                Viv.run(instance, Method("test", [map_x, 3]))
        print(result)  // 105.0

        Args:
            parameters (Any): VivJson's codes, JSON values,
                              file paths, variables,
                              Parsed object, or a Configuration
                            str: A VivJson's code, a JSON value, or
                                 a file path
                            list[Any] or tuple[Any]: Some VivJson's
                                                codes, JSON values,
                                                some file paths,
                                                variables, or
                                                Parsed objects
                            Json: JSON object
                            dict: Some variables that are name/value
                                  pairs.
                            Parsed: The parsed object that is generated
                                    with this method. It is passed
                                    through toward output.
                            Config: Configuration

                            Even if the class instance is contained,
                            it is ignored.
        Returns:
            Instance or NoneType, str: Class instance and Error message
                                1st value: Class instance if success,
                                           None otherwise.
                                2nd value: "" if success,
                                           Error message otherwise.
        """
        # Multiple parameters are represented as tuple.
        # | Caller of this method             | parameters            |
        # |-----------------------------------|-----------------------|
        # | make_instance("x = 3")            | ("x = 3",)            |
        # | make_instance("x = 3", "y = 5")   | ("x = 3", "y = 5")    |
        # | make_instance(["x = 3", "y = 5"]) | (["x = 3", "y = 5"],) |
        parsed, config, error_message = Viv._parse_internal(parameters)
        if error_message != '':
            return None, error_message

        if parsed is None or not isinstance(parsed.statements, list):
            return None, Viv.report_error('Invalid statement/value')
        block = Block(parsed.statements, Block.CLASS_CONSTRUCTOR)

        evaluator = Evaluator(config)
        try:
            evaluator.evaluate(block)
        except Error.EvaluateError as e:
            return None, str(e)
        except Exception as e:
            return None, Viv.report_error(str(e))
        return Instance(evaluator), ''

    @staticmethod
    def _run_internal(parsed, config=None):
        """Run VivJson with statements.

        Args:
            parsed (Parsed): Parsed object
            config (Config or NoneType, optional): Configuration if
                                                   needed,
                                                   None otherwise.

        Returns:
            Any, str: Returned value of script and Error message.
                      1st value: Returned value if success,
                                 None otherwise. (None may be
                                 returned even if success.)
                      2nd value: "" if success,
                                 Error message otherwise.
        """
        if parsed is None or not isinstance(parsed.statements, list):
            return None, Viv.report_error('Invalid statement/value')
        block = Block(parsed.statements, Block.ANONYMOUS_FUNCTION)

        evaluator = None
        if isinstance(parsed.instance, Instance) \
                and isinstance(parsed.instance.evaluator, Evaluator):
            evaluator = parsed.instance.evaluator
        else:
            evaluator = Evaluator(config)

        result = None
        try:
            result = evaluator.evaluate(block)
        except Error.EvaluateError as e:
            evaluator.rewind_after_abort()
            return None, str(e)
        except Exception as e:
            evaluator.rewind_after_abort()
            return None, Viv.report_error(str(e))
        return result, ''

    @staticmethod
    def print_statements(statements, add_class_name=False, config=None):
        """Print statements.

        Args:
            statements (list[Statement] or Block): Statements of script
            add_class_name (bool, optional): Class name is added to
                                        each statement if this is True.
                                        The default is False.
            config (Config or NoneType, optional): Configuration if
                                                   needed,
                                                   None otherwise.
        """
        if isinstance(statements, Block):
            statements = statements.values
        if not isinstance(statements, list):
            return

        for statement in statements:
            if isinstance(statement, Block):
                Viv.print_statements(statement)
                continue
            if isinstance(statement, Blank):
                continue

            string = Viv.make_string(statement, config=config)

            if isinstance(add_class_name, bool) and add_class_name:
                string = f'{statement.__class__.__name__}({string})'

            print(string)

    @staticmethod
    def make_string(value, config=None):
        """Convert a value into string. Serialize a value into JSON string.

        Configuration is available for Infinity and NaN (Not a Number).
        When it is not setting, "" (empty) is returned.

        Args:
            value (Any): Value
            config (Config or NoneType, optional): Configuration if
                                                   needed,
                                                   None otherwise.

        Returns:
            str: Converted value
        """
        string = ''
        try:
            string = Evaluator.make_str(value, config=config)
        except (Error.EvaluateError, Exception):
            pass
        return string

    @staticmethod
    def report_error(error_message: str, enable_stderr=False) -> str:
        """Reports Error.
   
        The formatted error is returned.
        The formatted error contains Tag looks like [xxxx].
        It is outputted into stderr when the argument "enableStderr" is
        true. 

        Args:
            error_message (str): Error message
            enable_stderr (bool, optional): Enablement of stderr. When
                                            it is true, the formatted
                                            error is outputted into
                                            stderr.
                                            The default is False.

        Returns:
            str: The formatted error
        """
        config = Config(enable_stderr=enable_stderr)
        return Error.report("", error_message, config=config)

    @staticmethod
    def _get_file_extension(word):
        """Get file extension if the given word is file path.

        Args:
            word (str): The file path or code

        Returns:
            str or NoneType: a file extension (one of the EXTENSIONS)
                             if the given word is file path,
                             None otherwise
        """
        if not isinstance(word, str):
            return None
        _, ext = os.path.splitext(word)
        ext = ext.lower()
        if ext not in Viv.EXTENSIONS:
            return None
        if not os.path.exists(word):
            return None
        return ext

    @staticmethod
    def _get_argument_text(index: int) -> str:
        """Get text message of argument index.

        Args:
            index (int): Argument index

        Returns:
            str: Text message of the given argument index
        """
        suffixes = ('st', 'nd', 'rd')
        suffix = suffixes[index] if index < len(suffixes) else 'th'
        index += 1
        return f'{index}{suffix} argument'

    @staticmethod
    def _extract_parameters(parameters):  # pylint: disable=R0912
        """Extract parameters.

        Args:
            parameters (list or tuple): VivJson's codes, JSON values,
                              file paths, variables,
                              Parsed objects, a class instance,
                              a method object, or a Configuration.

        Returns:
            list or NoneType, Instance or NoneType, Config or NoneType:
                        Extracted parameters and Configuration.
                        1st value: Extracted parameters if success,
                                   None otherwise.
                        2nd value: Class instance if it is given,
                                   None otherwise
                        3rd value: Configuration if it is given,
                                   None otherwise
        """
        instance = None
        for parameter in parameters:
            if isinstance(parameter, Instance):
                instance = parameter
                break

        extracted_parameters = []
        config = None
        for parameter in parameters:  # pylint: disable=R1702
            if isinstance(parameter, Instance):
                pass
            elif isinstance(parameter, (list, tuple)):
                extracted_parameters.extend(parameter)
            elif isinstance(parameter, (str, dict, Json, Parsed)):
                extracted_parameters.append(parameter)
            elif isinstance(parameter, Config):
                config = parameter
            elif isinstance(parameter, Method):
                ret = None
                if instance is not None and instance.evaluator is not None \
                        and isinstance(parameter.name, str) \
                        and parameter.name[0] != "_" \
                        and isinstance(parameter.arguments, (list, tuple)):
                    try:
                        value = instance.evaluator.get(parameter.name)
                        if isinstance(value, CalleeRegistry):
                            call = Viv._make_call(parameter)
                            if call is not None:
                                ret = Return(Token(Token.Type.RETURN), call)
                    except Error.EvaluateError:
                        pass
                if ret is None:
                    return None, None, config
                extracted_parameters.append(Parsed([ret], None))
            else:
                return None, None, config
        return extracted_parameters, instance, config

    @staticmethod
    def _concatenate_parameters(parameters):
        """Concatenate if + option is given.

        Args:
            parameters (list): Parameters that may be have "+".

        Returns:
            list[tuple] or NoneType, str: Concatenated parameters and
                                    "" if success,
                                    None and Error message otherwise.
                                    Each concatenated parameter is
                                    a tuple that has index (0-) and 
                                    parameter.
        """
        concatenated = []  # list[tuple], each tuple has index and parameter.
        ahead = None  # NoneType or list[tuple]
        for index, parameter in enumerate(parameters):
            if parameter == '+':
                try:
                    ahead = concatenated.pop()
                except IndexError:
                    return None, Viv.report_error(
                                            'No statement before + option')
                continue
            if ahead is not None:
                if not isinstance(parameter, str):
                    return None, Viv.report_error('Bad data after + option')
                index = ahead[0]
                parameter = ahead[1] + parameter
                ahead = None
            concatenated.append((index, parameter))
        if ahead is not None:
            return None, Viv.report_error('No statement after + option')
        return concatenated, ''

    @staticmethod
    def _make_call(method):
        """Make Call statement.

        Args:
            method (Method): Calling class method.
                             This object has 2 members.
                             "name" is class method name.
                             "arguments" is its arguments.

        Returns:
            Call or NoneType: Call statement if success, None otherwise
        """
        name = Identifier(Token(Token.Type.IDENTIFIER, lexeme=method.name))

        values = []
        for index, argument in enumerate(method.arguments):
            if not Viv._is_valid_type(argument):
                return None
            value = Value(argument, Viv._get_argument_text(index))
            values.append(value)
        return Call(name, Array(values))

    @staticmethod
    def _collect_only_valid_elements(value):
        """Collect only valid elements.

        Valid type: None, bool, int, float, str, list, dict

        Note that tuple is invalid.
        Note that key of dict must be str.

        Args:
            value (Any): Value

        Returns:
            Any: value if success, Environment.UNDEFINED otherwise
        """
        if value is None or isinstance(value, (bool, int, float, str)):
            return value
        if isinstance(value, list):
            fixed = []
            for element in value:
                element = Viv._collect_only_valid_elements(element)
                if element != Environment.UNDEFINED:
                    fixed.append(element)
            return fixed
        if isinstance(value, dict):
            fixed = {}
            for k, v in value.items():
                if not isinstance(k, str):
                    continue
                v = Viv._collect_only_valid_elements(v)
                if v != Environment.UNDEFINED:
                    fixed[k] = v
            return fixed
        return Environment.UNDEFINED

    @staticmethod
    def _is_valid_type(value):  # pylint: disable=R0911
        """Judge whether the given value is valid or not.

        Valid type: None, bool, int, float, str, list, dict

        Note that tuple is invalid.
        Note that key of dict must be str.

        Args:
            value (Any): Value

        Returns:
            bool: True if valid, False otherwise.
        """
        if value is None or isinstance(value, (bool, int, float, str)):
            return True
        if isinstance(value, list):
            for element in value:
                is_valid = Viv._is_valid_type(element)
                if not is_valid:
                    return False
            return True
        if isinstance(value, dict):
            for k, v in value.items():
                if not isinstance(k, str):
                    return False
                is_valid = Viv._is_valid_type(v)
                if not is_valid:
                    return False
            return True
        return False

    @staticmethod
    def _is_setting_implicit_variable(statements):
        """Judge whether it is setting implicit variable or not.

        The implicit variable is "_". Its setting is "_ = ...".

        Args:
            statements (list[Statement]): Statements of script

        Returns:
            bool: True if setting implicit variable,
                  False otherwise.
        """
        return isinstance(statements, list) and len(statements) == 1 \
                and isinstance(statements[0], Set) \
                and isinstance(statements[0].members[0], Identifier) \
                and statements[0].members[0].name.lexeme == '_'

    @staticmethod
    def _fix_setting_implicit_variable(statements, index):
        """Fix implicit variable.

        The original implicit variable is "_". It is fixed as
        "_[index]".

        Args:
            statements (list[Statement]): Statements of script
            index (int): The index of implicit variable.
                         It starts from 0.
        """
        selector = Literal(Token(Token.Type.NUMBER, lexeme=str(index)))
        statements[0].members.insert(1, selector)

@dataclass(frozen=True)
class Json:
    """JSON data class"""
    value: str

@dataclass(frozen=True)
class Instance:
    """Instance data class"""
    evaluator: Evaluator

METHODARG = TypeVar("METHODARG", list, tuple)

@dataclass(frozen=True)
class Method:
    """Method data class"""
    name: str
    arguments: METHODARG

@dataclass(frozen=True)
class Parsed:
    """Parsed data class"""
    statements: list[Statement]
    instance: Optional[Instance]



"""Unit-test for Command-line

Run `python3 -m pytest tests/test_commandline.py` in parent directory.

Environment:
- Python 3.9 or later
- pytest 8.3 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-26'

import subprocess

def test_commandline_basic():
    """Unit-test for basic Command-line"""
    cp = subprocess.run(['python3', '-m', 'vivjson', 'return(10)'],
                        encoding='utf-8', stdout=subprocess.PIPE,
                        check=False)
    result = cp.stdout.strip()
    assert result == '10'

    cp = subprocess.run(['python3', '-m', 'vivjson',
                         '3', '2', 'return(_[0] + _[1])'],
                        encoding='utf-8', stdout=subprocess.PIPE,
                        check=False)
    result = cp.stdout.strip()
    assert result == '5'

    cp = subprocess.run(['python3', '-m', 'vivjson',
                         '3', '2', 'return(_)'],
                        encoding='utf-8', stdout=subprocess.PIPE,
                        check=False)
    result = cp.stdout.strip()
    assert result == '[3, 2]'

    cp = subprocess.run(['python3', '-m', 'vivjson', '5', 'return(_ * 2)'],
                        encoding='utf-8', stdout=subprocess.PIPE,
                        check=False)
    result = cp.stdout.strip()
    assert result == '10'

    cp = subprocess.run(['python3', '-m', 'vivjson',
                         'for(i in [1,2,3]){print(i*5)};return("")'],
                        encoding='utf-8', stdout=subprocess.PIPE,
                        check=False)
    result = cp.stdout.strip()
    assert '5' in result
    assert '10' in result
    assert '15' in result

    cp = subprocess.run(['python3', '-m', 'vivjson', 'tests/call_6.viv'],
                        encoding='utf-8', stdout=subprocess.PIPE,
                        check=False)
    result = cp.stdout.strip()
    assert result == '6'

    cp = subprocess.run(['python3', '-m', 'vivjson',
                         'tests/a5b7c9.json', 'tests/axb-c.viv'],
                        encoding='utf-8', stdout=subprocess.PIPE,
                        check=False)
    result = cp.stdout.strip()
    assert result == '26'

    cp = subprocess.run(['python3', '-m', 'vivjson',
                         'tests/a5b7c9.json', 'return(a+b+c)'],
                        encoding='utf-8', stdout=subprocess.PIPE,
                        check=False)
    result = cp.stdout.strip()
    assert result == '21'

    cp = subprocess.run(['python3', '-m', 'vivjson',
                         'tests/dog2cat3.json',
                         'sum = 0; for (a in _) {sum += a.number}; return(sum)'],
                        encoding='utf-8', stdout=subprocess.PIPE,
                        check=False)
    result = cp.stdout.strip()
    assert result == '5'

def test_commandline_usage():
    """Unit-test for usage"""
    cp = subprocess.run(['python3', '-m', 'vivjson', 'a=3/0'],
                        encoding='utf-8', stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE, check=False)
    error = cp.stderr.strip()
    assert 'Error: Cannot evaluate' in error

    for option in [None, '-help', '--invalid']:
        command = ['python3', '-m', 'vivjson']
        if option is not None:
            command.append(option)

        cp = subprocess.run(command,
                            encoding='utf-8', stdout=subprocess.PIPE,
                            check=False)
        result = cp.stdout.strip()
        assert 'file extension' in result
        assert 'Example 1' in result

def test_commandline_stdin():
    """Unit-test for stdin"""
    for option in ['-i', '--stdin']:
        for stdin in ['a=3', '{"a": 3}']:
            command = \
                f'echo "{stdin}" | python3 -m vivjson {option} "return(a*2)"'
            cp = subprocess.run(command,
                                encoding='utf-8', stdout=subprocess.PIPE,
                                shell=True, check=False)
            result = cp.stdout.strip()
            assert result == '6'

    command = 'echo "3" | python3 -m vivjson -i=a "return(a+2)"'
    cp = subprocess.run(command,
                        encoding='utf-8', stdout=subprocess.PIPE,
                        shell=True, check=False)
    result = cp.stdout.strip()
    assert result == '5'

    command = 'echo "3" | python3 -m vivjson -i="" "return(a+2)"'
    cp = subprocess.run(command,
                        encoding='utf-8', stdout=subprocess.PIPE,
                        shell=True, check=False)
    result = cp.stdout.strip()
    assert result == '2'

    command = 'echo "3" | python3 -m vivjson -i= "return(a+2)"'
    cp = subprocess.run(command,
                        encoding='utf-8', stdout=subprocess.PIPE,
                        shell=True, check=False)
    result = cp.stdout.strip()
    assert result == '2'

    for option in ['-i', '--stdin']:
        for quotation in ['"', "'", ""]:
            command = 'cat tests/dog2cat3.json | ' \
                f'python3 -m vivjson {option}={quotation}values{quotation} ' \
                '"result = {}, ' \
                'for(value in values){result[value.name] = value.number}, ' \
                'return(result)"'
            cp = subprocess.run(command,
                                encoding='utf-8', stdout=subprocess.PIPE,
                                shell=True, check=False)
            result = cp.stdout.strip()
            assert result == '{"dog": 2, "cat": 3}'

    cp = subprocess.run('cat tests/call_6.viv | python3 -m vivjson -i',
                        encoding='utf-8', stdout=subprocess.PIPE,
                        shell=True, check=False)
    result = cp.stdout.strip()
    assert result == '6'

    cp = subprocess.run('python3 -m vivjson -i tests/call_6.viv',
                        encoding='utf-8', stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE, shell=True, check=False)
    result = cp.stdout.strip()
    assert result == ''
    error = cp.stderr.strip()
    assert 'Error: Cannot read from stdin.' in error

    command = 'cat tests/array_escaped_str.json | ' \
              'python3 -m vivjson -i=x "return(x)"'
    cp = subprocess.run(command,
                        encoding='utf-8', stdout=subprocess.PIPE,
                        shell=True, check=False)
    result = cp.stdout.strip()
    # Note that U+2028 (LINE SEPARATOR) and U+2029 (PARAGRAPH SEPARATOR) are
    # included.
    assert result == '["あいうえおか", "x\\ny", "x\\by", "x y", "x y"]'

    command = 'cat tests/array_escaped_str.json | ' \
              'python3 -m vivjson -i=x "print(x), return(\'\')"'
    cp = subprocess.run(command,
                        encoding='utf-8', stdout=subprocess.PIPE,
                        shell=True, check=False)
    result = cp.stdout.strip()
    # Note that U+2028 (LINE SEPARATOR) and U+2029 (PARAGRAPH SEPARATOR) are
    # included.
    assert result == '["あいうえおか", "x\\ny", "x\\by", "x y", "x y"]'

    command = 'cat tests/array_escaped_str.json | ' \
              'python3 -m vivjson -i=x "print(x[0], x[1], x[2], x[3], x[4]), return(\'\')"'
    cp = subprocess.run(command,
                        encoding='utf-8', stdout=subprocess.PIPE,
                        shell=True, check=False)
    result = cp.stdout.strip()
    # Note that U+2028 (LINE SEPARATOR) and U+2029 (PARAGRAPH SEPARATOR) are
    # included.
    assert result == 'あいうえおか, x\ny, x\by, x y, x y'

def test_commandline_version():
    """Unit-test for version"""
    for option in ['-v', '--version']:
        command = ['python3', '-m', 'vivjson']
        command.append(option)

        cp = subprocess.run(command,
                            encoding='utf-8', stdout=subprocess.PIPE,
                            check=False)
        result = cp.stdout.strip()
        assert 'specification version' in result
        assert 'interpreter version' in result



"""Unit-test for Viv class

Unit-test for "run", "parse_text", ...

Run `python3 -m pytest tests/test_viv.py` in parent directory.

Environment:
- Python 3.9 or later
- pytest 8.3 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-27'

# pylint: disable=C0301, E0401

import datetime
from math import isnan
import pytest
from vivjson.viv import Viv, Json, Method
from vivjson.config import Config
from vivjson.statement import Literal, Return
from vivjson.tokens import Token

def test_run():  # pylint: disable=R0915
    """Unit-test for "Viv.run" method."""
    value, error_message = Viv.run("3", "2", "return(_[0] + _[1])")
    assert error_message == ''
    assert value == 5

    value, error_message = Viv.run("3", "2", "return(_)")
    assert error_message == ''
    assert value == [3, 2]

    value, error_message = Viv.run("5", "return(_*2)")
    assert error_message == ''
    assert value == 10

    value, error_message = Viv.run("a:3,b:2,return(a+b)")
    assert error_message == ''
    assert value == 5

    value, error_message = Viv.run(["{a:3,b:2}", "{return(a+b)}"])
    assert error_message == ''
    assert value == 5

    value, error_message = \
        Viv.run(("x=", "+", "{a:3,b:2}", "return(x.a+x.b)"))
    assert error_message == ''
    assert value == 5

    value, error_message = Viv.run('{"foo": 3, "bar": 2}', "return(foo + bar)")
    assert error_message == ''
    assert value == 5
    value, error_message = Viv.run('{"foo": 3, "bar": 2}', "return('foo' in .)")
    assert error_message == ''
    assert value is True
    value, error_message = Viv.run('{"foo": 3, "bar": 2}', "return(qux)")
    assert error_message == ''
    assert value is None  # null
    value, error_message = Viv.run('{"foo": 3, "bar": 2}', "return('qux' in .)")
    assert error_message == ''
    assert value is False

    data = '{"foo": 10, "bar": 30, "baz": 20}'
    value, error_message = Viv.run(data)
    assert error_message == ''
    assert value == {"foo": 10, "bar": 30, "baz": 20}
    # print(value)  # {'foo': 10, 'bar': 30, 'baz': 20}
    # print(type(value))  # <class 'dict'>

    value, error_message = Viv.run(data, 'return(foo)')
    assert error_message == ''
    assert value == 10
    # print(value)  # 10
    # print(type(value))  # <class 'int'>

    value, error_message = Viv.run(data, 'return(qux)')
    assert error_message == ''
    assert value is None
    # print(value)  # None
    # print(type(value))  # <class 'NoneType'>

    value, error_message = Viv.run(data, 'return("qux" in .)')
    assert error_message == ''
    assert value is False
    # print(value)  # False
    # print(type(value))  # <class 'bool'>

    value, error_message = Viv.run(data, 'return(foo + bar + baz)')
    assert error_message == ''
    assert value == 60
    # print(value)  # 60

    code = 'max=-1, for (pair in pairs) {if (max < pair[1]) {max = pair[1]}}, return(max)'
    value, error_message = Viv.run('pairs = ', '+', data, code)
    assert error_message == ''
    assert value == 30
    # print(value)  # 30
    value, error_message = Viv.run('pairs = ' + data, code)
    assert error_message == ''
    assert value == 30
    # print(value)  # 30
    code = 'max=-1, for (pair in .) {if (max < pair[1]) {max = pair[1]}}, return(max)'
    value, error_message = Viv.run(data, code)
    assert error_message == ''
    assert value == 30
    # print(value)  # 30
    code = '_max=-1, for (pair in .) {if (_max < pair[1]) {_max = pair[1]}}, return(_max)'
    value, error_message = Viv.run(data, code)
    assert error_message == ''
    assert value == 30
    # print(value)  # 30

    value, error_message = Viv.run()
    assert error_message == ''
    assert value == {}

    value, error_message = Viv.run("x=", 3)
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run(("x=", "+"))
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run("x=", "+", "+", "+")
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run("x=", "+", Config(max_depth=3))
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run("x=", "+", {"a": 3})
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run({"___a___": 3})
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run(("+", "{a:3,b:2}"))
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run("tests/call_6.viv")
    assert error_message == ''
    assert value == 6

    value, error_message = Viv.run("tests/empty.viv")
    assert error_message == ''
    assert value == {}

    value, error_message = Viv.run("tests/broken.viv")
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run("tests/dummy.viv")
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run("tests/call_6.vev")
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run("")
    assert error_message == ''
    assert value == {}

    value, error_message = Viv.run("tests/a5b7c9.json", "tests/axb-c.viv")
    assert error_message == ''
    assert value == 26

    value, error_message = Viv.run("tests/a5b7c9.json", "return(a+b+c)")
    assert error_message == ''
    assert value == 21

    value, error_message = Viv.run("tests/dog2cat3.json",
                                       "sum = 0; for (a in _) {sum += a.number}; return(sum)")
    assert error_message == ''
    assert value == 5

    value, error_message = Viv.run({"x":[1,2],"y":True},"return(if(y){:=x[0]}else{:=x[1]})")
    assert error_message == ''
    assert value == 1

    value, error_message = Viv.run({"x":[1,2],"y":False},"return(if(y){:=x[0]}else{:=x[1]})")
    assert error_message == ''
    assert value == 2

    value, error_message = Viv.run({"x":[1,2]},{"y":True},"return(if(y){:=x[0]}else{:=x[1]})")
    assert error_message == ''
    assert value == 1

    value, error_message = Viv.run({"x":[1,2]},{"y":False},"return(if(y){:=x[0]}else{:=x[1]})")
    assert error_message == ''
    assert value == 2

    value, error_message = Viv.run({"x":1},"y: x", {"x":10}, "y += x, return(y)")
    assert error_message == ''
    assert value == 11

    value, error_message = Viv.run({"x": {"y": 1}}, "return(x.y)")
    assert error_message == ''
    assert value == 1

    value, error_message = Viv.run({"x": [1]}, "return(x.0)")
    assert error_message == ''
    assert value == 1

    value, error_message = Viv.run({"x": {"0": 1}}, "return(x.0)")
    assert error_message == ''
    assert value == 1

    value, error_message = Viv.run({"x": {0: 1}}, "return(x.0)")
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run({"x": Literal(Token(Token.Type.NUMBER, lexeme="100"))}, "return(x)")
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run({"x":datetime.datetime.now(),"y":False},"return(if(y){:=x.year}else{:=x.month})")
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run("a = [1, 2, 3, 4, 5]", Config(max_array_size=4))
    assert error_message != ''
    assert value is None

def test_parse_and_run():
    """Unit-test for "Viv.parse" and "Viv.run" method."""
    parsed, error_message = \
        Viv.parse("3", "2", "return(_[0] + _[1])", Config(max_depth=3))
    assert error_message == ''
    assert parsed is not None

    value, error_message = Viv.run(parsed)
    assert error_message == ''
    assert value == 5

    parsed, error_message = \
        Viv.parse({"x":[1,2]},{"y":True},"return(if(y){:=x[0]}else{:=x[1]})")
    assert error_message == ''
    assert parsed is not None

    value, error_message = Viv.run(parsed)
    assert error_message == ''
    assert value == 1

    parsed, error_message = Viv.parse("return(foo + bar)")
    assert error_message == ''
    assert parsed is not None
    value, error_message = Viv.run('{"foo":3, "bar": 2}', parsed)
    assert error_message == ''
    assert value == 5
    # print(value);  # 5

def test_parse_file_and_run():
    """Unit-test for "Viv.parse_file" and "Viv.run" method."""
    parsed, error_message = Viv.parse_file("tests/call_6.viv")
    assert error_message == ''
    assert parsed is not None

    value, error_message = Viv.run(parsed)
    assert error_message == ''
    assert value == 6

    parsed, error_message = Viv.parse_file(5)
    assert error_message != ''
    assert parsed is None

    parsed, error_message = Viv.parse_file("tests/empty.viv")
    assert error_message == ''
    assert parsed is not None

    value, error_message = Viv.run(parsed)
    assert error_message == ''
    assert value == {}

    parsed, error_message = Viv.parse_file("tests/invalid.txt")
    assert error_message != ''
    assert parsed is None

def test_parse_text_and_run():
    """Unit-test for "Viv.parse_text" and "Viv.run" method."""
    parsed, error_message = Viv.parse_text("return(3 + 2)")
    assert error_message == ''
    assert parsed is not None

    value, error_message = Viv.run(parsed)
    assert error_message == ''
    assert value == 5

    parsed, error_message = Viv.parse_text("return(3")
    assert error_message != ''
    assert parsed is None

    parsed, error_message = Viv.parse_text(3)
    assert error_message != ''
    assert parsed is None

def test_run_with_variables():
    """Unit-test for "Viv.run" method."""
    parsed, error_message = Viv.parse_text("return(x.a+x.b)")
    assert error_message == ''
    assert isinstance(parsed.statements, list)
    assert isinstance(parsed.statements[0], Return)
    text = str(parsed.statements[0])
    assert text == 'return(x["a"] + x["b"])'

    variable = {"x": {"a": 70, "b": 30}}

    value, error_message = Viv.run(variable, parsed)
    assert error_message == ''
    assert value == 100

    value, error_message = Viv.run(1, parsed)
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run(variable, 3)
    assert error_message != ''
    assert value is None

    parsed, error_message = \
        Viv.parse_text("return(if(y){:=x[0]}else{:=x[1]})")
    assert error_message == ''
    variable = {"x":[1,2],"y":True}
    value, error_message = Viv.run(variable, parsed)
    assert error_message == ''
    assert value == 1

    variable["y"] = False
    value, error_message = Viv.run(variable, parsed)
    assert error_message == ''
    assert value == 2

    parsed, error_message = \
        Viv.parse("y: x", {"x":10}, "y += x, return(y)")
    assert error_message == ''
    assert parsed is not None

    value, error_message = Viv.run({"x": 100}, parsed)
    assert error_message == ''
    assert value == 110

    parsed, error_message = Viv.parse_text("return(x.0)")
    assert error_message == ''
    assert isinstance(parsed.statements, list)

    value, error_message = Viv.run({"x": [100]}, parsed)
    assert error_message == ''
    assert value == 100

    value, error_message = Viv.run({"x": {"0": 100}}, parsed)
    assert error_message == ''
    assert value == 100

    value, error_message = Viv.run({"x": {0: 100}}, parsed)
    assert error_message != ''
    assert value is None

TEST_MAKE_INSTANCE_AND_RUN_DATA = (
    ([3, 8], 11, False),
    ([3, None], 3, False),
    ([True, False], True, False),
    (['foo', 'bar'], 'foobar', False),
    ([['foo', 'bar'], 'baz'], ['foo', 'bar', 'baz'], False),
    ([('foo', 'bar'), 'baz'], None, True),
    ((('foo', 'bar'), 'baz'), None, True),
    ([{'foo': 2, 'bar': 5}, {'baz': 10}], {'foo': 2, 'bar': 5, 'baz': 10}, False),
    ([{'foo': 2, 'bar': 5}, {1: 10}], None, True),
    ([1, [datetime.datetime.now(), 2]], None, True),
    ([{'foo': 2, 'bar': 5}, {'baz': datetime.datetime.now()}], None, True),
)

def test_make_instance_and_run():  # pylint: disable=R0915
    """Unit-test for "Viv.make_instance" and "Viv.run" method."""
    code = '_x = 3 ' \
           'function add(a, b) { ' \
           '  return(a + b) ' \
           '} ' \
           'y = true'

    instance, error_message = Viv.make_instance(code)
    assert error_message == ''
    assert instance is not None

    for data in TEST_MAKE_INSTANCE_AND_RUN_DATA:
        result, error_message = Viv.run(instance, Method('add', data[0]))
        assert (error_message == '') != data[2]
        assert result == data[1]

    parsed, error_message = Viv.parse(code)
    assert error_message == ''
    assert parsed is not None
    instance, error_message = Viv.make_instance(parsed)
    assert error_message == ''
    assert instance is not None
    for data in TEST_MAKE_INSTANCE_AND_RUN_DATA:
        result, error_message = Viv.run(instance, Method('add', data[0]))
        assert (error_message == '') != data[2]
        assert result == data[1]

    result, error_message = Viv.run(instance, Method('add', [100, -10]))
    assert error_message == ''
    assert result == 90

    result, error_message = Viv.run(Method('add', [100, -10]), instance)
    assert error_message == ''
    assert result == 90

    instance, error_message = Viv.make_instance(
        'function add(a, b) { '
        '  return(a + b) '
        '} '
        'y = 3 / 0'
    )
    assert error_message != ''
    assert instance is None

    instance, error_message = Viv.make_instance(
        'function div(a, b) { '
        '  return(a / b) '
        '} '
        'function _div(a, b) { '
        '  return(a / b) '
        '} '
        'y = 3 / 2'
    )
    assert error_message == ''
    assert instance is not None
    result, error_message = Viv.run(instance, Method('div', [10, 0]))
    assert error_message != ''
    assert result is None
    result, error_message = Viv.run(instance, Method('div', [10, 2]))
    assert error_message == ''
    assert result == 5
    result, error_message = Viv.run(instance, Method('div', (10, 2)))
    assert error_message == ''
    assert result == 5
    result, error_message = Viv.run(instance, Method('_div', [10, 2]))
    assert error_message != ''
    assert result is None
    result, error_message = Viv.run(datetime.datetime.now(), Method('div', [10, 10]))
    assert error_message != ''
    assert result is None
    result, error_message = Viv.run(instance, Method('add', [3, 2]))
    assert error_message != ''
    assert result is None
    result, error_message = Viv.run(instance, Method(1, [3, 2]))
    assert error_message != ''
    assert result is None
    result, error_message = Viv.run(instance, Method('div', {"a": 3, "y": 2}))
    assert error_message != ''
    assert result is None

    instance, error_message = Viv.make_instance(True)
    assert error_message != ''
    assert instance is None

    code = "function test(x, y) {" \
            "  z = x.a + x.b.1 + y" \
            "  return(z)" \
            "}"
    map_x = {"a": 100, "b": [1.0, 2.0]}
    instance, error_message = Viv.make_instance(code)
    assert error_message == ''
    assert instance is not None
    result, error_message = Viv.run(instance, Method("test", [map_x, 3]))
    assert error_message == ''
    assert result == 105.0

    data = '["foo", 10, [{"bar": null, "baz": "test"}, false]]'
    instance, error_message = Viv.make_instance(data)
    assert error_message == ''
    assert instance is not None
    value, error_message = Viv.run(instance, "return(_[2][0]['bar'])")
    assert error_message == ''
    assert value is None  # null
    # print(value)  # None (null is represented as None in Python.)
    value, error_message = Viv.run(instance, "return(_.2.-2.baz)")
    assert error_message == ''
    assert value == 'test'
    # print(value)  # test

    code = 'function add(a, b) {' \
           '  return(a + b)' \
           '}'
    instance, error_message = Viv.make_instance(code)
    assert error_message == ''
    assert instance is not None
    value, error_message = Viv.run(instance, Method('add', [10, 20]))
    assert error_message == ''
    assert value == 30
    # print(value)  # 30
    value, error_message = Viv.run(instance, 'return(add(10, 20))')
    assert error_message == ''
    assert value == 30
    # print(value)  # 30

    code = "function add(a, b) {" \
           "  return(a + b)" \
           "}" \
           "c = [20, false]"
    instance, error_message = Viv.make_instance(code)
    assert error_message == ''
    assert instance is not None
    value, error_message = Viv.run(instance, '{"foo":3, "bar": 2}', 'return(add(foo, bar))')
    assert error_message == ''
    assert value == 5
    # print(value)  # 5
    value, error_message = Viv.run(instance, Method("add", [3, 2]))
    assert error_message == ''
    assert value == 5
    # print(value)  # 5

    value, error_message = Viv.run(instance, "return(c[0])")
    assert error_message == ''
    assert value == 20
    # print(value)  # 20

@pytest.mark.parametrize(
    'inf, nan, prepared_expression, target_expression, expected_value, is_error', (
        ('Infinity', 'NaN', '"a": {"x": Infinity}', 'a.x', float('inf'), False),
        ('Infinity', 'NaN', '"a": {"x": Infinity}', 'a.x + 1', float('inf'), False),
        ('Infinity', 'NaN', '"a": {"x": -Infinity}', 'a.x', float('-inf'), False),
        ('Infinity', 'NaN', '"a": {"x": -Infinity}', 'a.x + 1', float('-inf'), False),
        ('Infinity', 'NaN', '"a": {"x": NaN}', 'a.x', float('nan'), False),
        ('Infinity', 'NaN', '"a": {"x": NaN}', 'a.x + 1', float('nan'), False),
        (None, None, '"a": {"x": Infinity}', 'a.x', None, False),
        (None, None, '"a": {"x": Infinity}', 'a.x + 1', 1, False),
        (None, None, '"a": {"x": -Infinity}', 'a.x', None, False),
        (None, None, '"a": {"x": -Infinity}', 'a.x + 1', 1, False),
        (None, None, '"a": {"x": NaN}', 'a.x', None, False),
        (None, None, '"a": {"x": NaN}', 'a.x + 1', 1, False),
    )
)

def test_run_inf_nan(inf, nan, prepared_expression, target_expression, expected_value, is_error):  # pylint: disable=R0913
    """Unit-test for "Viv.run" method and infinity/nan"""
    config = Config(infinity=inf, nan=nan)
    statement = f'{prepared_expression}; return({target_expression})'
    value, error_message = Viv.run(statement, config)
    assert (error_message != '') == is_error
    if isinstance(expected_value, float) and isnan(expected_value):
        assert isnan(value) is True
    else:
        assert value == expected_value
    if inf is not None and nan is not None and not is_error:
        value, error_message = Viv.run(prepared_expression, config)
        text = Viv.make_string(value, config)
        assert text == f'{{{prepared_expression}}}'

def test_run_inf_nan_with_file():
    """Unit-test for "Viv.run" method and infinity/nan with a JSON file."""
    value, error_message = Viv.run('tests/inf_nan.json')
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run('tests/inf_nan.json', Config(infinity="Infinity", nan="NaN"))
    assert error_message == ''
    assert isinstance(value, dict)
    assert value["normal"] == 1.5
    assert value["inf"] == float('inf')
    assert value["negative_inf"] == float('-inf')
    assert isnan(value["nan"])
    assert value["str"] == "1.5"

def test_json():  # pylint: disable=R0915
    """Unit-test for JSON."""
    text1 = "{\"a\": 5, \"b\": [2, 1]}"
    # Valid as Script's code and JSON's value
    value, error_message = Viv.run(text1)
    assert error_message == ''
    assert Viv.make_string(value) == text1
    # Valid as JSON's value
    value, error_message = Viv.run(Json(text1))
    assert error_message == ''
    assert Viv.make_string(value) == text1

    text2 = "{\"a\": 3 + 2, \"b\": [2, 1]}"
    # Valid as Script's code and JSON's value
    value, error_message = Viv.run(text2)
    assert error_message == ''
    assert Viv.make_string(value) == text1
    # Invalid as JSON's value: 3 + 2
    value, error_message = Viv.run(Json(text2))
    assert error_message != ''  # Error
    assert value is None  # Result is nothing

    code = "return(a)"
    text1 = "{\"a\": 5, \"b\": [2, 1]}"
    # Valid as Script's code and JSON's value
    value, error_message = Viv.run(text1, code)
    assert error_message == ''
    assert value == 5
    # Valid as JSON's value
    value, error_message = Viv.run(Json(text1), code)
    assert error_message == ''
    assert value == 5

    text2 = "{\"a\": 3 + 2, \"b\": [2, 1]}"
    # Valid as Script's code and JSON's value
    value, error_message = Viv.run(text2, code)
    assert error_message == ''
    assert value == 5
    # Invalid as JSON's value: 3 + 2
    value, error_message = Viv.run(Json(text2), code)
    assert error_message != ''  # Error
    assert value is None  # Result is nothing

    data = '{"foo": 3}'
    code = 'return(foo)'
    value, error_message = Viv.run(Json(data), code)
    assert error_message == ''
    assert value == 3
    # print(value)  # 3

    code = "return(_)"
    text1 = "5"
    # Valid as Script's code and JSON's value
    value, error_message = Viv.run(text1, code)
    assert error_message == ''
    assert value == 5
    # Valid as JSON's value
    value, error_message = Viv.run(Json(text1), code)
    assert error_message == ''
    assert value == 5

    text2 = "10 / 2"
    # Valid as Script's code and JSON's value
    value, error_message = Viv.run(text2, code)
    assert error_message == ''
    assert value == 5
    # Invalid as JSON's value: 10 / 2
    value, error_message = Viv.run(Json(text2), code)
    assert error_message != ''  # Error
    assert value is None  # Result is nothing

    code = "return(a)"
    text1 = "{\"a\": Infinity}"
    # *** With configuration ***
    config = Config()
    config.set_infinity("Infinity")
    # Valid as Script's code and JSON's value
    value, error_message = Viv.run(text1, code, config)
    assert error_message == ''
    assert value == float('inf')
    # Valid as JSON's value
    value, error_message = Viv.run(Json(text1), code, config)
    assert error_message == ''
    assert value == float('inf')

    # *** Without configuration ***
    # Valid as Script's code and JSON's value
    # However, Infinity is treated as the undefined variable.
    # So the evaluated result will be null.
    value, error_message = Viv.run(text1, code)
    assert error_message == ''
    assert value is None  # None
    # Invalid as JSON's value because the value of key-value is not
    # number, string, boolean, object, array, or null.
    # The above Infinity is not string because it is not surrounded
    # with quotation marks.
    value, error_message = Viv.run(Json(text1), code)
    assert error_message != ''  # Error
    assert value is None  # Result is nothing

    code = "return(_)"
    text1 = "Infinity"
    # *** With configuration ***
    config = Config()
    config.set_infinity("Infinity")
    # Valid as Script's code and JSON's value
    value, error_message = Viv.run(text1, code, config)
    assert error_message == ''
    assert value == float('inf')
    # Valid as JSON's value
    value, error_message = Viv.run(Json(text1), code, config)
    assert error_message == ''
    assert value == float('inf')

    # *** Without configuration ***
    # Valid as Script's code and JSON's value
    # However, it is treated as String.
    value, error_message = Viv.run(text1, code)
    assert error_message == ''
    assert value == text1
    # Valid as JSON's value
    # However, it is treated as String.
    value, error_message = Viv.run(Json(text1), code)
    assert error_message == ''
    assert value == text1

@pytest.mark.parametrize(
    'value, expected_string, inf, nan', (
        (2, '2', None, None),
        (1.5, '1.5', None, None),
        (1.0, '1.0', None, None),
        (float('inf'), '', None, None),
        (float('inf'), 'Infinity', 'Infinity', None),
        (float('inf'), '', None, 'NaN'),
        (float('inf'), 'Infinity', 'Infinity', 'NaN'),
        (float('-inf'), '', None, None),
        (float('-inf'), '-Infinity', 'Infinity', None),
        (float('-inf'), '', None, 'NaN'),
        (float('-inf'), '-Infinity', 'Infinity', 'NaN'),
        (float('nan'), '', None, None),
        (float('nan'), '', 'Infinity', None),
        (float('nan'), 'NaN', None, 'NaN'),
        (float('nan'), 'NaN', 'Infinity', 'NaN'),
        (True, 'true', None, None),
        (False, 'false', None, None),
        ("abcd", 'abcd', None, None),
        (None, 'null', None, None),
        ([], '[]', None, None),
        ([1,2], '[1, 2]', None, None),
        ((), '[]', None, None),
        (('a','b'), '["a", "b"]', None, None),
        ({}, '{}', None, None),
        ({'a': 3.0, 'b': True}, '{"a": 3.0, "b": true}', None, None),
        ({'a': 3.0, 'b': {"c": [True, None]}}, '{"a": 3.0, "b": {"c": [true, null]}}', None, None),
    )
)

def test_make_string(value, expected_string, inf, nan):
    """Unit-test for "Viv.make_string" method."""
    config = Config(infinity=inf, nan=nan)
    string = Viv.make_string(value, config=config)
    assert string == expected_string



"""Environment for VivJson

- Environment: Constructor.
- Environment#get
    - Get a variable's value if argument is variable name.
    - Get a block if argument is empty.
      A block has all variable without variables of the prefix "_"
      and definitions of pure function or closure.
    - Get a result value that is given with ":=" statement.
      It is returned if argument is empty and a result value is
      registered.
- Environment#set: Set a variable.
- Environment#remove: Remove a variable.

The following types are stored.

- Host native literal (str, int, float, bool, None)
- Host native array (list)
- Host native block (dict)
- Definition of pure function or closure (CalleeRegistry)
- Variable whose value is standard library's function (Identifier)
- Alias (Get)

Refer to:
- "Crafting Interpreters"
  https://craftinginterpreters.com/
Note that this code is made from scratch. The source code
of the above WEB site is not used.

Environment:
- Python 3.9 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-29'

from .statement import CalleeRegistry

class Environment:
    """Environment class

    Attributes:
        _enclosing (Environment or NoneType): The parent environment.
                                              None is given when this
                                              instance is root.
        _variables (dict): Variable name and value are stored.
        UNDEFINED (tuple): It is used to represent undefined variable.
        RESULT_NAME (str): It is used to assign particular value as
                           result by ':='.
        FUNCTION_RETURN_NAME (str): It is used for "return" statement.
                            Firstly, when anonymous/pure function
                            starts, it creates this variable whose
                            value is UNDEFINED.
                            Secondly, "return" statement sets value
                            and evaluation goes back until finishing
                            function. When the returned value is
                            given like "return(xxxx)", its value is
                            set. Otherwise, this name is set as
                            value.
                            For example, the function "test" creates
                            this variable. Then "return 'zero'" sets
                            "zero" into this variable and evaluation
                            goes back to assignment of "x".
                              function test(a) {
                                  if (a == 0) {
                                      return 'zero'
                                  }
                                  :
                                  :
                              }

                              x = test(0)
        BREAK_NAME (str): It is used for "break" statement.
                          Firstly, when loop starts, it creates this
                          variable whose value is UNDEFINED.
                          Secondly, "break" statement sets any value
                          and evaluation goes back until finishing
                          loop.
        CONTINUE_NAME (str):  It is used for "continue" statement.
                          Firstly, when loop starts, it creates this
                          variable whose value is UNDEFINED.
                          Secondly, "continue" statement sets any
                          value and evaluation goes back until
                          starting loop.
    """
    UNDEFINED = ()
    RESULT_NAME = '___#RESULT#___'
    FUNCTION_RETURN_NAME = '___#RETURN#___'
    BREAK_NAME = '___#BREAK#___'
    CONTINUE_NAME = '___#CONTINUE#___'

    def __init__(self, enclosing=None):
        """Initialize class.

        Args:
            enclosing (Environment or NoneType, optional): The parent
                                environment. None is given when this
                                instance is root.
                                The default is None.
        """
        self._enclosing = enclosing
        self._variables = {}

    def get(self, name=None, only_this_scope=False):
        """Get variable's value.

        Get variable's value.
        When it is not existed in this scope, it is tried in the parent
        scope.
        None is returned if it is not existed in the whole scope.

        When name is None, the whole values are returned as dict.
        However it is excluded when the prefix of variable is "_" or
        value is the definition of function (CalleeRegistry).
        Note that a result value of ":=" statement had been registered,
        it is returned.

        Args:
            name (str, optional): Variable name.
                                  When this is None, the whole values
                                  are returned as dict. However it is
                                  excluded when the prefix of variable
                                  is "_" or value is the function
                                  (CalleeRegistry).
                                  Note that a result value of ":="
                                  statement had been registered, it is
                                  returned.
                                  The default is None.
            only_this_scope (bool, optional): When it is True, getting
                                            scope is limited only here.
                                            The default is False.

        Returns:
            Any: Its value.
                 Environment.UNDEFINED is returned if the given name's
                 variable is not existed.
        """
        if name is None:
            # :=
            if Environment.RESULT_NAME in self._variables:
                return self._variables[Environment.RESULT_NAME]

            # Whole values
            result = {}
            for key, value in self._variables.items():
                if key[0] == '_':
                    continue
                if isinstance(value, CalleeRegistry) \
                        and not value.is_reference:
                    continue
                result[key] = value
            return result

        # Try to get from current scope.
        if not isinstance(name, str):
            return Environment.UNDEFINED
        if name in self._variables:
            return self._variables[name]

        # Try to get from parent scope.
        if not only_this_scope and self._enclosing is not None:
            return self._enclosing.get(name)

        return Environment.UNDEFINED

    def set(self, name, value, only_this_scope=False):
        """Set a variable.

        Modify/Assign a variable.
        Firstly, modifying is tried. When the given variable is not
        existed in this scope, it is tried in the parent scope.
        When it is not existed in the whole scope, it is assigned in
        this scope newly.

        Args:
            name (str or NoneType): Variable name.
                                    When None is given, value is set
                                    as the returned value.
            value (Any): Its value
            only_this_scope (bool, optional): When it is True, setting
                                            scope is limited only here.
                                            The default is False.
        """
        if name is None:
            name = Environment.RESULT_NAME  # :=
        elif not isinstance(name, str):
            return

        if not only_this_scope:
            is_completed = self.modify(name, value)
            if is_completed:
                return
        self._variables[name] = value  # New assignment

    def modify(self, name, value):
        """Modify a variable.

        Modify a variable.
        When the given variable is not existed in this scope, it is
        tried in the parent scope.

        Args:
            name (str): Variable name
            value (Any): Its value

        Returns:
            bool: True if modifying is completed,
                  False otherwise.
        """
        if name in self._variables:
            self._variables[name] = value  # Modify
            return True

        if self._enclosing is not None:
            return self._enclosing.modify(name, value)

        return False

    def remove(self, name, only_this_scope=False):
        """Remove a variable.

        Args:
            name (str): Variable name.
            only_this_scope (bool, optional): When it is True, removing
                                            scope is limited only here.
                                            The default is False.

        Returns:
            bool: True if removing is completed,
                  False otherwise.
        """
        if name in self._variables:
            del self._variables[name]
            return True

        if not only_this_scope and self._enclosing is not None:
            return self._enclosing.remove(name)

        return False

    def get_enclosing(self):
        """Get enclosing.

        Returns:
            Environment or NoneType: enclosing that is the parent
                                     Environment.
                                     None if enclosing is not
                                     available. In other words,
                                     this is root Environment
                                     if None is returned.
        """
        return self._enclosing



"""Unit-test for Parser

Unit-test for Parser with Lexer

Run `python3 -m pytest tests/test_parser.py` in parent directory.

Environment:
- Python 3.9 or later
- pytest 8.3 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-04-01'

# pylint: disable=C0301, E0401, R0801, C0123

import pytest
from vivjson.viv import Viv
from vivjson.config import Config

@pytest.mark.parametrize(
    'prepared_expression, target_expression, expected_value, is_error', (
        ('a = 1.5', 'a', 1.5, False),
        ('a= 1.5e1', 'a', 15.0, False),
        ('a = 1.', 'a', None, True),  # Digit is needed after "."
        ('a =', 'a', None, True),  # Missing value
        ('a = {}, a.1 = {}, a.1.2 = true', 'a', {"1": {"2": True}}, False),
        ('a = {}, a.1/2 = {}', 'a', None, True),  # Invalid member
        ('[a, b] = [1, 3]', 'a', None, True),  # Invalid assignment
        ('a + 1 = 1', 'a', None, True),  # Invalid assignment
        ('a: ""', 'a', "", False),
        ('a = "text\'', 'a', None, True),  # Different quotation mark is invalid
        ('a = "text\\" b = 3', 'a', None, True),  # Missing end of "
        ('a = "text\\"', 'a', None, True),  # Missing end of "
        ('a = "text\\\\text"', 'a', 'text\\text', False),
        ('a = "text\ttext"', 'a', 'text\ttext', False),
        ('a = "\\u3042\\u3044\\u3046\\u3048\\u304a\\u304B"', 'a', 'あいうえおか', False),
        ('a = "\\u4a\\u3044"', 'a', None, True),  # Invalid unicode
        ('a = "\\u4a"', 'a', None, True),  # Invalid unicode
        ('a = "\\u"', 'a', None, True),  # Invalid unicode
        ('a = "x\\"y"', 'a', 'x"y', False),
        ('a = "x\\\'y"', 'a', "x'y", False),
        ('a = "x\\\\y"', 'a', "x\\y", False),
        ('a = "x\\/y"', 'a', "x/y", False),
        ('a = "x\\by"', 'a', "x\by", False),
        ('a = "x\\fy"', 'a', "x\fy", False),
        ('a = "x\\ny"', 'a', "x\ny", False),
        ('a = "x\\ry"', 'a', "x\ry", False),
        ('a = "x\\ty"', 'a', "x\ty", False),
        ('a : true', 'a', True, False),
        ('a := 1.5', 'a', None, True),  # := is invalid after Identifier.
        ('x = if (true) {:=10}', 'x', 10, False),
        ('x = if (true) {a:=10}', 'x', None, True),  # := is invalid after Identifier.
        ('x = if (true) {:=}', 'x', None, True),  # Missing value
        ('a = [1 5 {b: {c: 10} d: -20}] e = -100', 'a[-1].b.c + e', -90, False),
        ('a=[1 5 {b:{c:10} d:-20}] e=-100', 'a[-1].b.c+e', -90, False),
        ('a=[1 5 {b:{c:10} d:-20}] e=-100', 'a.-1.b.c+e', -90, False),
        ('a=[1 5 {b:{c:10} d:-20}] e=-100', 'a.-2', 5, False),
        ('a=[1 5 {b:{c:10} d:-20}] e=-100', 'a.-2.', None, True),  # It is invalid that a dot is the tail.
        ('a=[1 5 {b:{c:10} d:-20}] e=-100', 'a . \n -2', 5, False),
        ('a=[1 5 {b:{c:10} d:-20}] e=-100', 'a.2.b.c+e', -90, False),
        ('a=[1 5 {b:{c:10} d:-20}] e=-100', 'a.2. b.c+e', -90, False),
        ('a=[1 5 {b:{c:10} d:-20}] e=-100', 'a . 2 . b.c+e', -90, False),
        ('a\n=\n[1\n\n{b:{c:10} d:-20}]\ne\n\n=\n-100', 'a[-1]\n. \nb.c+e', -90, False),
        ('', '-3+1e-4*1000', -2.9, False),
        ('a = [5]', 'a[0', None, True), # It is invalid that the right bracket is lacked.
        ('{a = [5 }', '3', None, True), # It is invalid that the right bracket is lacked.
        ('', '3 *+ 4', 12, False),
        ('', '3 +* 4', None, True),  # +* is invalid
        ('', '(3 + 4 {1 + 2}', None, True),  # It is invalid that the right bracket is lacked.
        ('', '() + 2', None, True),  # It is invalid that there is no expression in group.
        ('a := 2', 'a', None, True),
        ('', '3 > 2 > -1', None, True),  # It should be fixed as 3 > 2 and 2 > -1.
        ('a = not false', 'a', True, False),
        ('a = not(false)', 'a', True, False),
        ('a = true and not false', 'a', True, False),
        ('a = !false', 'a', None, True),  # ! is invalid
        ('function x2(a) {return(a * 2)}', 'x2(4)', 8, False),
        ('function x2(a){return(a*2)}', 'x2(4+2)', 12, False),
        ('function x2() {}', 'x2(1)', {}, False),
        ('function x2() a = 3', 'a', None, True),  # Function needs block (statements).
        ('function test(a + b) {}', 'test()', None, True),  # Invalid parameter
        ('function test(break a) {}', 'test()', None, True),  # Invalid modifier
        ('function run(function worker) {worker()}', 'a', None, False),
        ('function run(function) {return}', 'a', None, True),  # Identifier is needed after argument's modifier.
        ('function (a) {return}', 'a', None, True),  # Identifier is needed after function's modifier.
        ('a = 0\n\nif\n\n(\n\ntrue\n\n)\n\n{\n\na = 3\n\n}', 'a', 3, False),
        ('a = 0 if (true) {}', 'a', 0, False),
        ('if () { a = 3 }', 'a', None, True),  # Condition is necessary.
        ('if (true), a = 3', 'a', None, True),  # Block (operations) is necessary.
        ('if (true, a = 3)', 'a', None, True),  # Block (operations) is necessary.
        ('while () {i=0}', 'i', None, True),  # Condition is necessary.
        ('while (true)', 'i', None, True),  # Block (operations) is necessary.
        ('while {i = 0}', 'i', None, True),  # Condition (parenthesis) is necessary.
        ('for {i = 0}', 'i', None, True),  # Condition (parenthesis) is necessary.
        ('a = 0, for (i = 0; i < 20; i += 1; { a = 2 })', 'a', 2, False),
        ('for (i = 0; i < 20; i += 1; { a = 2 }', 'i', None, True),  # Both parentheses are necessary.
        ('for (i = 0; i < 20; i += 1)', 'i', None, True),  # Block (operations) is necessary.
        ('for (i = 0; i < 20; i += 1, a = 3)', 'a', None, True),  # Block (operations) is necessary.
        ('for ()', 'i', None, True),  # Block (operations) is necessary.
        ('for (3 in [2, 3]) {break}', 'i', None, True),  # Iterator needs Identifier for left-hand side.
        ('sum = 0, for (i = 5, i < 7, i += 1) {sum += i}', 'sum', 11, False),
        ('sum = 0, for (i = 5, i < 7, i += 1, i += 1) {sum += i}', 'sum', None, True),  # Number of argument is invalid.
        ('remove, a = 3', 'a', None, True),  # remove function needs argument.
        ('remove(), a = 3', 'a', None, True),  # remove function needs argument.
        ('remove(a', 'a', None, True),  # Missing ")"
        ('a = 3,a *= 5;a += 2\na = a - 27 a /= 4\tb = 3   a+=b', 'a', 0.5, False),
        ('a = 20, a = 30 # ', 'a', 30, False),
        ('a = 20 # , a = 30', 'a', 20, False),
        ('a = 20 // , a = 30', 'a', 20, False),
        ('a = /* 20, a = */ 30', 'a', 30, False),
        ('a = /* 20, a = 30', 'a', None, True),  # Missing "*/"
        ('a = 20 # abc\na = 30', 'a', 30, False),
        ('a /* = 20\na */ = 30', 'a', 30, False),
        ('a = 20 /* \na = 30 */', 'a', 20, False),
        ('a = [\n\n1\n\n2\n\n3\n\n]', 'a', [1, 2, 3], False),
        ('a = [\n1,\n2,\n3\n]', 'a', [1, 2, 3], False),
        ('a = [\n1,\n,\n3\n]', 'a', None, True),  # Missing 2nd value though delimiter is existed
        ('function test(\na\nb\nc\n)\n{return(a+b+c)}\nx=test(10, 100, 1000)', 'x', 1110, False),
        ('function test(\n\na\n#comment\nb\n\nc\n\n)\n{return(a+b+c)}\nx=test(10, 100, 1000)', 'x', 1110, False),
        ('function test(a, , c)\n{return(a+b+c)}\nx=test(10, 100, 1000)', 'x', None, True),  # Missing 2nd parameter though delimiter is existed
        ('function test(a, b, c)\n{return(a+b+c)}\nx=test(10, , 1000)' , 'x', 1010, False),  # Missing 2nd argument will be null.
    )
)

def test_parser(prepared_expression, target_expression, expected_value,
                is_error):
    """Unit-test for Parser

    Unit-test for Parser with Lexer

    Args:
        prepared_expression (str or dict): When the another expression,
                                 such as assignment, is needed before
                                 test, it is given here.
                                 Maybe this is '' mostly.
                                 str: Statements as VivJson
                                 dict: Value as Python's value
        target_expression (str): "return" method's argument as targeted
                                 expression.
        expected_value (Any): The expected value as Python's value
        is_error (bool): True if occurring error is expected,
                         False otherwise.
    """
    config = Config(max_depth=11)

    text = f'return({target_expression})'
    value, error_message = Viv.run(prepared_expression, text, config)
    assert (error_message != '') == is_error

    if isinstance(value, float):
        value = round(value, 10)
    assert value == expected_value

    if error_message == '' \
            and not isinstance(expected_value, bool) \
            and isinstance(expected_value, (int, float)):
        assert type(value) == type(expected_value)  # pylint: disable=C0123

@pytest.mark.parametrize(
    'is_only_json, expression, expected_value, is_error', (
        (True, '{"a": 3, "b": [true, null, 1.5e3]}', {"a": 3, "b": [True, None, 1500.0]}, False),
        (False, '{"a": 3, "b": [true, null, 1.5e3]}', {"a": 3, "b": [True, None, 1500.0]}, False),

        (True, '{"a": 3, "b": true}', {"a": 3, "b": True}, False),
        (False, '{"a": 3, "b": true}', {"a": 3, "b": True}, False),

        (True, '"a": -3, "b": true', {"a": -3, "b": True}, False),
        (False, '"a": -3, "b": true', {"a": -3, "b": True}, False),

        (True, '"a": 3, "b": true, return(a)', None, True),
        (False, '"a": 3, "b": true, return(a)', 3, False),

        (True, '{"a": 3 + 2, "b": true}', None, True),
        (False, '{"a": 3 + 2, "b": true}', {"a": 5, "b": True}, False),

        (True, '10', {}, False),
        (False, '10', {}, False),
    )
)

def test_json(is_only_json, expression, expected_value, is_error):
    """Unit-test of JSON for Parser

    Args:
        is_only_json (bool): is_only_json of Config
        expression (str): The targeted expression
        expected_value (Any): The expected value as Python's value
        is_error (bool): True if occurring error is expected,
                         False otherwise.
    """
    config = Config(enable_only_json=is_only_json)

    value, error_message = Viv.run(expression, config)
    assert (error_message != '') == is_error
    if isinstance(value, float):
        value = round(value, 10)
    assert value == expected_value
    assert type(value) == type(expected_value)

@pytest.mark.parametrize(
    'json, script, expected_value, is_error', (
        ('"a": 3, "b": true', 'return(a)', 3, False),
        ('"a": 3, "b": true', '', {"a": 3, "b": True}, False),
        ('"a": 3, "b": true', 'return("")', '', False),
        ('"a": 3, "b": true', 'return(a+2)', 5, False),
        ('"a": 3 + 2, "b": true', 'return(a)', None, True),  # "+" operator is not allowed in JSON.
        ('"a": 3, "b": true', 'a += 2, b = not b', {"a": 5, "b": False}, False),
        ('10', '', {}, False),  # The number(, string, boolean, null, array) can't be returned without script.
        ('10', 'return(_)', 10, False),
        ('false', '', {}, False),  # The number(, string, boolean, null, array) can't be returned without script.
        ('false', 'return(_)', False, False),
        ('test', '', {}, False),  # The number(, string, boolean, null, array) can't be returned without script.
        ('test', 'return(_)', 'test', False),
        ('null', '', {}, False),  # The number(, string, boolean, null, array) can't be returned without script.
        ('null', 'return(_)', None, False),
        ('[1, 2, 3]', '', {}, False),  # The number(, string, boolean, null, array) can't be returned without script.
        ('[1, 2, 3]', 'return(_)', [1, 2, 3], False),
    )
)

def test_json_and_script(json, script, expected_value, is_error):
    """Unit-test of JSON & Script for Parser

    Args:
        json (str): The targeted JSON value 
        expression (str): The targeted expression
        expected_value (Any): The expected value as Python's value
        is_error (bool): True if occurring error is expected,
                         False otherwise.
    """
    config = Config(enable_only_json=True)
    parsed, error_message = Viv.parse(json, config)
    assert (error_message != '') == is_error
    if error_message == '':
        assert parsed is not None
    else:
        assert parsed is None
    value, error_message = Viv.run(parsed, script)
    assert (error_message != '') == is_error
    if isinstance(value, float):
        value = round(value, 10)
    assert value == expected_value
    assert type(value) == type(expected_value)

def test_file():
    """Unit-test of file for Parser"""
    parsed, error_message = Viv.parse('tests/invalid_as_json.json')
    assert error_message != ''
    assert parsed is None
    parsed, error_message = Viv.parse('tests/invalid_as_json.viv')
    assert error_message == ''
    assert parsed is not None

    parsed, error_message = Viv.parse_file('tests/invalid_as_json.json')
    assert error_message != ''
    assert parsed is None
    parsed, error_message = Viv.parse_file('tests/invalid_as_json.viv')
    assert error_message == ''
    assert parsed is not None

    config = Config()
    assert config.get_enable_only_json() is False
    parsed, error_message = Viv.parse_file('tests/invalid_as_json.json', config)
    assert error_message != ''
    assert parsed is None
    assert config.get_enable_only_json() is False

    value, error_message = Viv.run('tests/array.viv')
    assert error_message == ''
    count = 0
    for v in value.values():
        assert v == [1, 2, 3]
        count += 1
    assert count == 8
